"""Unit test for preprocessing functions
"""
import numpy as np
import adafdr.method as md

def test_feature_preprocess():
    """ Test for method.feature_preprocess
    """
    x_test = np.array([[0.1, 3, 2, 5],
                       [0.1, 10, 5, 18]])
    x_test = md.feature_preprocess(x_test.T)
    print(x_test)
    assert all(x_test[:, 0] == x_test[:, 1])

def test_get_order_discrete():
    """ Test for method.get_order_discrete
    """
    np.random.seed(0)
    x_test = np.random.choice([0, 1, 2, 3], size=300)
    temp = np.arange(300)
    x_val_test = np.array([0, 1, 2, 3])
    p_test = np.ones([300], dtype=float)
    p_test[x_test == 0] = 0.001
    p_test[(x_test == 1)*(temp < 200)] = 0.001
    p_test[(x_test == 2)*(temp < 100)] = 0.001
    x_order = md.get_order_discrete(p_test, x_test, x_val_test)
    print(x_order)
    assert all(x_order == [3, 2, 1, 0])
    x_new_test = md.reorder_discrete(x_test, x_val_test, x_order)
    assert all((x_test == 0) == (x_new_test == 3))
    assert all((x_test == 1) == (x_new_test == 2))
    assert all((x_test == 2) == (x_new_test == 1))
    


import numpy as np
import scipy as sp
from scipy import stats
from sklearn.mixture import GaussianMixture
import torch
from torch.autograd import Variable
from adafdr.util import *
import multiprocessing as mp
import logging
import matplotlib.pyplot as plt
import torch.nn.functional as tf
import time

np.set_printoptions(precision=4,suppress=True)

""" 
    preprocessing: standardize the hypothesis features 
    
    ----- input  -----
    x_: np_array, n*d, the hypothesis features.
    qt_norm: bool, if perform quantile normalization.
    return_metainfo: bool, if return the meta information regarding the features
    vis_dim: list, the dimensions to visualize. counting starts from 0, needed only when verbose is True
    verbose: bool, if generate ancillary information
    
    ----- output -----
    x: Processed feature stored as an n*d array. The discrete feature is reordered based on the alt/null ratio.
    meta_info: a d*2 array. The first dimensional corresponds to the type of the feature (continuous/discrete). The second 
               dimension is a list on the mapping (from small to large). For example, if the before reorder is [0.1,0.2,0.7].
               This list may be [0.7,0.1,0.2].
""" 

def feature_preprocess(x, qt_norm=True, continous_rank=True):
    """Feature preprocessing: 1. quantile normalization, 
        2. make the range [0,1] for each feature

    Args:
        x ((n,d) ndarray): The covaraites.
        qt_norm (bool): If perform quantile normalization.
        continous_rank (bool): Indicate if break the same value by randomization.
        
    Returns:
        x ((n,d) ndarray): The discrete feature is reordered based on the alt/null ratio.
    """
    if len(x.shape) == 1: 
        x = x.reshape([-1,1])
    n,d = x.shape     
    if qt_norm:
        x=rank(x, continous_rank=continous_rank)       
    # Rescale to be within [0,1].
    x = (x-x.min(axis=0))/(x.max(axis=0)-x.min(axis=0))        
    return x

def get_feature_type(x):
    """ Tell if a feature is continuous or discrete.
    """
    if np.unique(x).shape[0]<100:
        return 'discrete'
    else: 
        return 'continuous'

def get_order_discrete(p, x, x_val, n_full=None):
    """ Calculate the order of the discrete features according to the alt/null ratio
    Args:
        p ((n,) ndarray): The p-values.
        x ((n,) ndarray): The covaraites. The data is assumed to have been preprocessed.
        x_val ((n_val,) ndarray): All possible values for x, sorted in ascending order.
        n_full (int): Total number of hypotheses before filtering.    
        
    Returns:
        x_order ((d,) ndarray): the order (of x_val) from smallest alt/null ratio to 
            the largest.
    """
    n_val = x_val.shape[0]
    # Separate the null and the alt proportion.
    _, t_BH = bh_test(p, alpha=0.1, n_full=n_full)
    x_null, x_alt = x[p>0.75], x[p<t_BH]      
    # Calculate the alt/null ratio。
    cts_null = np.zeros([n_val], dtype=int)
    cts_alt  = np.zeros([n_val], dtype=int)
    for i,val in enumerate(x_val):
        cts_null[i] = np.sum(x_null==val)+1
        cts_alt[i] = np.sum(x_alt==val)+1
    p_null  = cts_null/np.sum(cts_null)
    p_alt   = cts_alt/np.sum(cts_alt)      
    p_ratio = p_alt/p_null        
    # Calculate the order of x_val based on the ratio.
    x_order = p_ratio.argsort()
    return x_order

def reorder_discrete(x, x_val, x_order):
    """ Reorder the covariates based on the new order.
    Args:
        x ((n,) ndarray): The covaraites.
        x_val ((n_val,) ndarray): All possible values for x, sorted in ascending order.
        x_order ((n_val,) ndarray): The new order of x_val.
        
    Returns:
        x_new ((d,) ndarray): the reordered covaraites.
    """
    x_new = np.copy(x)
    n_val = x_val.shape[0]
    for i in range(n_val):
        x_new[x == x_val[x_order[i]]] = x_val[i]
    return x_new

def adafdr_explore(p, x_input, alpha=0.1, n_full=None, vis_dim=None,\
                           cate_name={}, output_folder=None,\
                           title_list=None, h=None):
    """Provide a visualization of pi1/pi0 for each dimension,
    to visualize the amount of information carried in each dimension.
    
        Args:
        p ((n,) ndarray): the p-values.
        x_input ((n,d) ndarray): The covaraites. The discrete features should be coded by
            integers starting from 0.
        alpha (float): The nominal FDR level.
        n_full (int): Total number of hypotheses before filtering.
        vis_dim (list or ndarray): the dimensions to visualize
        cate_name (dic of dics): the names of discrete categories for each dimension.
            None when the category names are not available. An example is as follows:
            cate_name = {1: {0: 'name0', 1: 'name1'}}. Here, dimension 0 has
            no names. Dimension 1 is discrete and has two values 0,1, whose corresponding
            names are name0, name1
        output_folder (string): The output directory.
        title_list (list of strings): Titles for each figure. 
        h ((n,) ndarray): The ground truth (None if not available). 
    
    Returns:
    """
    def plot_feature_1d(x_margin, p, x_null, x_alt, meta_info,\
                        title='',cate_name=None, output_folder=None,\
                        h=None, ):
        # Some parameters for generating the figures.
        feature_type,cate_order,x_val_ = meta_info       
        x_min,x_max = np.percentile(x_margin, [1,99])
        x_range = x_max-x_min
        x_min -= 0.01*x_range
        x_max += 0.01*x_range
        bins = np.linspace(x_min,x_max, 51)
        bin_width = bins[1]-bins[0]
        # Ploting for the continuous case and the discrete case.
        if feature_type == 'continuous':         
            # Continuous: use kde to estimate the probability.
            n_bin = bins.shape[0]-1
            x_grid = (bins+bin_width/2)[0:-1]
            p_null,_ = np.histogram(x_null, bins=bins) 
            p_alt,_ = np.histogram(x_alt, bins=bins)   
            p_null = p_null+1
            p_alt = p_alt+1
            p_null = p_null/np.sum(p_null)*n_bin
            p_alt = p_alt/np.sum(p_alt)*n_bin            
            kde_null = stats.gaussian_kde(x_null).evaluate(x_grid)
            kde_alt = stats.gaussian_kde(x_alt).evaluate(x_grid)
            psuedo_density = np.min(kde_null[kde_null>0])/10
            psuedo_density = psuedo_density.clip(min=1e-20)
            p_ratio = (kde_alt+psuedo_density)/(kde_null+psuedo_density)
            
        else: 
            # Discrete: use the empirical counts.
            unique_null,cts_null = np.unique(x_null, return_counts=True)
            unique_alt,cts_alt = np.unique(x_alt, return_counts=True)            
            unique_val = np.array(list(set(list(unique_null)+list(unique_alt))))
            unique_val = np.sort(unique_val)            
            p_null,p_alt = np.zeros([unique_val.shape[0]]), np.zeros([unique_val.shape[0]])          
            for i,key in enumerate(unique_null): 
                p_null[unique_val==key] = cts_null[i]                
            for i,key in enumerate(unique_alt): 
                p_alt[unique_val==key] = cts_alt[i]           
            n_bin = unique_val.shape[0]           
            p_null = (p_null+1)/np.sum(p_null+1)*n_bin
            p_alt = (p_alt+1)/np.sum(p_alt+1)*n_bin            
            p_ratio = p_alt/p_null  
            x_grid = (np.arange(unique_val.shape[0])+1)/(unique_val.shape[0]+1)
            x_min,x_max,bin_width = 0,1,1/(unique_val.shape[0]+1)
            if cate_name is None: 
                cate_name_ = x_val_[cate_order]
            else:
                cate_name_ = []  
                for i in cate_order:
                    cate_name_.append(cate_name[x_val_[i]])
        # Generate the figure.
        rnd_idx=np.random.permutation(p.shape[0])[0:np.min([10000, p.shape[0]])]
        if feature_type == 'continuous':
            pts_jitter = 0*(np.random.rand(10000)-0.5)
        else:
            pts_jitter = 0.75*(np.random.rand(10000)-0.5)
        plt.figure(figsize=[5, 3])
        p = p[rnd_idx]
        x_margin = x_margin[rnd_idx]
        # Ground truth.
        if h is not None:
            h = h[rnd_idx]
            plt.scatter(x_margin[h==1]+pts_jitter[h==1],\
                        p[h==1], color='orange', alpha=0.3, s=4, label='alt')
            plt.scatter(x_margin[h==0]+pts_jitter[h==0],\
                        p[h==0], color='steelblue', alpha=0.3, s=4, label='null')
        else:
            plt.scatter(x_margin+pts_jitter,\
                        p, color='steelblue', alpha=0.3, s=4, label='alt')
        if x_max>1e4:
            plt.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
        plt.ylim([0, min((p[p<0.5].max()), 0.05)])
        plt.ylabel('p-value', fontsize=14)
        plt.xlabel('covariate $x$', fontsize=14)
        if feature_type == 'discrete':
            plt.xticks(x_val_, cate_name_, rotation=90, fontsize=12)
        plt.tight_layout()
        if output_folder is not None:
            plt.savefig(output_folder+'/explore_p_%s.png'%title)
            plt.savefig(output_folder+'/explore_p_%s.pdf'%title)
            plt.close()
        else:
            plt.show()
        plt.figure(figsize=[5, 3])
        plt.bar(x_grid, p_null, width=bin_width, color='steelblue', alpha=0.6, label='null')
        plt.bar(x_grid, p_alt, width=bin_width, color='orange', alpha=0.6, label='alt')
        plt.xlim([x_min, x_max])
        if feature_type=='discrete': 
            plt.xticks(x_grid, cate_name_, rotation=90, fontsize=12)
        elif x_max>1e4:
            plt.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
        plt.ylabel('null/alt proportion', fontsize=14)
        plt.legend(fontsize=12)
        # plt.subplot(313)
        # if feature_type == 'continuous':
        #     plt.plot(x_grid, p_ratio, color='seagreen', label='ratio', linewidth=4) 
        # else:
        #     plt.plot(x_grid, p_ratio, color='seagreen', marker='o', label='ratio', linewidth=4)
        #     plt.xticks(x_grid, cate_name_, rotation=45)
        # plt.xlim([x_min, x_max])
        # y_min,y_max = plt.ylim()
        # plt.ylim([0, y_max])
        # plt.ylabel('ratio')
        plt.xlabel('covariate $x$', fontsize=14)
        plt.tight_layout()
        if output_folder is not None:
            plt.savefig(output_folder+'/explore_ratio_%s.png'%title)
            plt.savefig(output_folder+'/explore_ratio_%s.pdf'%title)
            plt.close()
        else:
            plt.show()
    # Start of the main function. Make a copy of the data.
    np.random.seed(0)
    x = x_input.copy()
    if n_full is None: 
        n_full = p.shape[0]
    if len(x.shape) == 1: 
        x = x.reshape([-1, 1])
    _,d = x.shape   
    # Reorder the discrete features.
    meta_info = []        
    for i in range(d):
        feature_type = get_feature_type(x[:, i])
        if feature_type == 'discrete':
            x_val = np.sort(np.unique(x[:, i]))
            x_order = get_order_discrete(p, x[:, i], x_val, n_full=n_full)
            x[:, i] = reorder_discrete(x[:, i], x_val, x_order)
            meta_info.append([feature_type, x_order, x_val])
        else:
            meta_info.append([feature_type, None, None])    
    # Separate the null proportion and the alternative proportion.
    _,t_BH = bh_test(p, n_full=n_full, alpha=0.1)
    x_null,x_alt = x[p>0.75],x[p<t_BH]      
    # Generate figures.
    if vis_dim is None: 
        vis_dim = np.arange(min(5, d))    
    for i in vis_dim:
        x_margin = x[:, i]
        temp_null,temp_alt = x_null[:, i], x_alt[:, i]
        if i in cate_name.keys():
            temp_cate_name = cate_name[i]
        else: 
            temp_cate_name = None
        if title_list is not None:
            temp_title = title_list[i]
        else:
            temp_title = 'feature_%s'%str(i+1)
        plot_feature_1d(x_margin, p, temp_null, temp_alt,\
                        meta_info[i], title=temp_title,\
                        cate_name=temp_cate_name, output_folder=output_folder, h=h)            

def preprocess_two_fold(p1, x1, x2, n_full, f_write):
    """Data preprocessing two folds of data. Note that to prevent overfitting,
    the preprocessing procedure does not depend on the p-values from the second fold.
    
    Args:
        p1 ((n1,) ndarray): the p-value from the first fold.
        x1 ((n1,d) ndarray): the covariates from the first fold.
        x2 ((n1,d) ndarray): the covariates from the second fold.
        n_full (int): Total number of hypotheses before filtering.
    
    Returns:
        x1 ((n1,d) ndarray): the preprocessed covariates from the first fold.
        x2 ((n1,d) ndarray): the preprocessed covariates from the second fold.
    """
    if len(x1.shape) == 1: 
        x1 = x1.reshape([-1,1])
        x2 = x2.reshape([-1,1])
    n1,d = x1.shape
    n2,d = x2.shape
    n = n1+n2
    x1_new = x1.copy()
    x2_new = x2.copy()
    # Rearrange the discrete features.
    for i in range(d):
        temp_x = np.concatenate([x1[:, i], x2[:, i]])
        feature_type = get_feature_type(temp_x)
        if feature_type == 'discrete':
            x_val = np.sort(np.unique(temp_x))           
            x_order = get_order_discrete(p1, x1[:, i], x_val, n_full=n_full)
            x1_new[:, i] = reorder_discrete(x1[:, i], x_val, x_order)
            x2_new[:, i] = reorder_discrete(x2[:, i], x_val, x_order) 
            if f_write is not None:
                f_write.write('# dim %d, x_order=%s\n'%(i, x_order))
    # Rescale the data to [0,1] and perform quantile normalization.
    x = np.zeros([n, d])
    rand_idx = np.random.permutation(n1+n2)    
    x[rand_idx[0:n1], :] = x1_new
    x[rand_idx[n1:], :] = x2_new  
    x = feature_preprocess(x, qt_norm=True)
    x1_new = x[rand_idx[0:n1], :]
    x2_new = x[rand_idx[n1:], :]
    return x1_new, x2_new
    
def method_single_fold_wrapper(data):
    """ A wrapper for method_single_fold to be called by method_cv
    
    Args:
        data (list): data[0]: p1, x1, 
            data[1]: p2, x2,
            data[2]: K,alpha,n_full,n_itr,output_folder,random_state,
            data[3]: Fold number.
        
    Returns:
        n_rej (int): Number of rejected hypotheses.
        t2 ((n/2,) ndarray): The thresholds on the second fold of data.
        [a,b,w,mu,sigma,gamma]: Learned (reparametrized) parameters.
    """
    # Load the data.
    p1,x1 = data[0]
    p2,x2 = data[1]
    K, alpha, n_full,n_itr, output_folder, random_state, verbose, fast_mode = data[2]
    fold_number = data[3]
    # Start a write file.
    if output_folder is not None:
        fname = output_folder+'/record_fold_%d.txt'%fold_number
        f_write = open(fname,'w+')
        f_write.write('### record for fold_%d\n'%fold_number)
        f_write.write('# K=%d, alpha=%0.2f, n_full=%d, n_itr=%d, random_state=%d\n'\
                      %(K,alpha,n_full,n_itr,random_state))
    else:
        f_write = None
    np.random.seed(random_state+fold_number)
    # Preprocess the data.
    x1, x2 = preprocess_two_fold(p1, x1, x2, n_full, f_write)
    # Learn the threshold
    _,_,theta = method_single_fold(p1, x1, K=K, alpha=alpha, n_full=n_full, n_itr=n_itr,\
                                   verbose=verbose, output_folder=output_folder,\
                                   fold_number=fold_number, random_state=random_state,\
                                   f_write=f_write, fast_mode=fast_mode)
    a,b,w,mu,sigma,gamma = theta
    # Test on the second fold
    t2 = t_cal(x2,a,b,w,mu,sigma)
    gamma = rescale_mirror(t2,p2,alpha,f_write=f_write,title='cv')
    t2 = gamma*t2
    if f_write is not None:
        f_write.write('\n## Test result with method_cv fold_%d\n'%fold_number)
        result_summary(p2<t2, f_write=f_write, title='method_single_fold_wrapper_%d'%fold_number)
        f_write.close()
    return np.sum(p2<t2), t2, [a,b,w,mu,sigma,gamma]

def adafdr_test(p_input, x_input, K=5, alpha=0.1, n_full=None, n_itr=1500, qt_norm=True,\
                h=None, verbose=False, output_folder=None, random_state=0,\
                single_core=True, fast_mode=True):
    """Hypothesis testing with hypothesis splitting.

    Args:
        p_input ((n,) ndarray): The p-values.
        x_input ((n,d) ndarray): The covaraites. The discrete features should be coded by integers
            starting from 0.
        K (int): The number of bump components.
        alpha (float): The nominal FDR level.
        n_full (int): Total number of hypotheses before filtering.
        n_itr (int): The number of iterations for the optimization process.
        qt_norm (bool): If perform quantile normalization.
        h ((n,) ndarray): The ground truth (None if not available).        
        verbose (bool): Indicate if output the computation details.
        output_folder (string): The output directory.
        random_state (int): The random seed.
        single_core (bool): True: use single core. False: process two processes in parallel.
        fast_mode (bool): If True, go without optimization.
        
    Returns:
        n_rej (list): Number of rejected hypothesesfor the two folds.
        t ((n,) ndarray): The decision threshold.
        theta (list): Learned (reparametrized) parameters with the format [a,b,w,mu,sigma,gamma].
    """
    np.random.seed(random_state)    
    p = np.copy(p_input)
    x = np.copy(x_input)
    if fast_mode:
        single_core = False
    if n_full is None:
        n_full = p.shape[0]
    start_time=time.time()
    if len(x.shape) == 1: 
        x = x.reshape([-1,1])
    n_sample,d = x.shape
    n_sub = int(n_sample/2)
    # Randomly split the hypotheses into two folds.    
    rand_idx = np.random.permutation(n_sample)    
    fold_idx = np.zeros([n_sample],dtype=int)
    fold_idx[rand_idx[0:n_sub]] = 0
    fold_idx[rand_idx[n_sub:]] = 1
    # Construct the input data.
    args = [K, alpha, int(n_full/2), n_itr, output_folder, random_state, verbose, fast_mode]
    data_fold_1 = [p[fold_idx==0], x[fold_idx==0]]
    data_fold_2 = [p[fold_idx==1], x[fold_idx==1]]
    Y_input = []
    Y_input.append([data_fold_2, data_fold_1, args, 1])
    Y_input.append([data_fold_1, data_fold_2, args, 2])
    # Parallel processing the testing of the two folds.
    if single_core:
        res = list()
        res.append(method_single_fold_wrapper(Y_input[0]))
        res.append(method_single_fold_wrapper(Y_input[1]))
    else:
        po = mp.Pool(2)
        res = po.map(method_single_fold_wrapper, Y_input)
    # Summarize the result.
    n_rej = [res[0][0], res[1][0]]
    t = np.zeros([n_sample], dtype=float)
    t[fold_idx==0] = res[0][1]
    t[fold_idx==1] = res[1][1]
    theta = [res[0][2], res[1][2]]    
    if verbose:
        print('# total rejection: %d'%np.array(n_rej).sum(), n_rej)
        if h is not None:          
            print('# D=%d, FD=%d, FDP=%0.3f'\
                  %(np.sum(p<t), np.sum((p<t)*(h==0)),\
                    np.sum((p<t)*(h==0))/np.sum(p<t)))        
        # Visualize the learned threshold.
        if output_folder is not None:
            color_list = ['steelblue', 'orange']            
            n_figure = min(d,5)
            plt.figure(figsize=[8, 4+n_figure])
            for i_dim in range(n_figure):
                plt.subplot(str(n_figure)+'1'+str(i_dim+1))
                for i in range(2):
                    if h is not None:
                        plot_scatter_t(t[fold_idx==i], p[fold_idx==i], x[fold_idx==i, i_dim],\
                                       h[fold_idx==i], color=color_list[i], label='fold %d'%(i+1))
                    else:
                        plot_scatter_t(t[fold_idx==i], p[fold_idx==i], x[fold_idx==i, i_dim],\
                                       color=color_list[i], label='fold %d'%(i+1))
                plt.legend()
                plt.xlabel('x_%d'%(i_dim+1))
                plt.ylabel('p-value')
            plt.savefig(output_folder+'/learned_threshold.png')
            plt.close()
        print('#time total: %0.4fs'%(time.time()-start_time))
    res_adafdr = {'n_rej':n_rej, 'decision':p_input<=t, 'threshold':t, 'theta':theta}
    return res_adafdr
    # return (p_input<=t), t, theta

def method_single_fold(p_input, x_input, K=5, alpha=0.1, n_full=None,\
                       n_itr=1500, h=None, verbose=False,\
                       output_folder=None, fold_number=0, random_state=0, f_write=None,\
                       fast_mode=False):
    """Learn the decision threshold via optimization.

    Args:
        p_input ((n,) ndarray): The p-values.
        x_input ((n,d) ndarray): The covaraites. The data is assumed to have been preprocessed.
        K (int): The number of bump components.
        alpha (float): The nominal FDR level.
        n_full (int): Total number of hypotheses before filtering.
        n_itr (int): The number of iterations for the optimization process.
        h ((n,) ndarray): The ground truth (None if not available).        
        verbose (bool): Indicate if output the computation details.
        output_folder (string): The output directory.
        random_state (int): The random seed.
        fold_number (int(0,1)): The fold number.
        f_write (file handler (write mode)): The output file.
        fast_mode (bool): If True, go without optimization.
        
    Returns:
        n_rej (int): Number of rejected hypotheses.
        t ((n,) ndarray): The decision threshold.
        theta (list): Learned (reparametrized) parameters with the format [a,b,w,mu,sigma,gamma].
    """
    torch.manual_seed(random_state)
    p = np.copy(p_input)
    x = np.copy(x_input)
    if len(x.shape)==1:
        x = x.reshape([-1,1])
    d = x.shape[1]
    if f_write is not None:
        f_write.write('# n_sample=%d\n'%(x.shape[0]))
        for i in [1e-6,1e-5,1e-4,5e-4]:
            f_write.write('# p<%0.6f: %d\n'%(i,np.sum(p<i)))
        f_write.write('\n')    
    # Threshold initialization using method_init.
    a_init, mu_init, sigma_init, w_init = method_init(p, x, K, alpha=alpha,
                                                      n_full=n_full, verbose=verbose,\
                                                      output_folder=output_folder,\
                                                      random_state=random_state,\
                                                      fold_number=fold_number,\
                                                      f_write=f_write)
    # Reparametrization.
    a,b,w,mu,sigma = reparametrize(a_init, mu_init, sigma_init, w_init, d)
    # Rescale using mirror estiamtor. 
    t = t_cal(x,a,b,w,mu,sigma)    
    gamma = rescale_mirror(t,p,alpha,f_write=f_write,title='before optimization')
    t = gamma*t
    b,w = b+np.log(gamma),w+np.log(gamma)
    # Return the result without optimization.
    if fast_mode or (np.sum(p<t)<100):
        if f_write is not None:
            f_write.write('\n# Less than 100 discoveries. Switch to fast mode\n')
        n_rej = np.sum(p < t)
        theta = [a,b,w,mu,sigma,gamma]
        return n_rej, t, theta
    # Setting parameters for the optimization.
    # lambda0: adaptively set based on the approximation accuracy of the sigmoid function.
    lambda0,n_rej,n_fd = 1/t.mean(),max(np.sum(p<t),1),np.sum(p>1-t)
    if f_write is not None:
        f_write.write('\n## choosing lambda0\n')
        f_write.write('## lambda0=%0.1f, D=%d, FD_hat=%d, alpha_hat=%0.3f\n'%(lambda0,n_rej,n_fd,n_fd/n_rej))        
    while np.absolute(np.sum(sigmoid((lambda0*(t-p))))-n_rej)>0.02*n_rej \
        or np.absolute(np.sum(sigmoid((lambda0*(t+p-1))))-n_fd)>0.02*n_fd:
        lambda0 = lambda0+0.5/t.mean()
        if f_write is not None:
            f_write.write('# lambda0=%0.1f, D_apr=%0.1f (r err=%0.3f), FD_hat_apr=%0.3f (r err=%0.3f)\n'\
                          %(lambda0,np.sum(sigmoid((lambda0*(t-p)))),np.absolute(1-np.sum(sigmoid((lambda0*(t-p))))/n_rej),\
                            np.sum(sigmoid((lambda0*(t+p-1)))),np.absolute(1-np.sum(sigmoid((lambda0*(t+p-1))))/n_fd)))   
    # Other parameters. 
    lambda1  = 10/alpha
    lambda0, lambda1 = int(lambda0), int(lambda1)
    loss_rec = np.zeros([n_itr],dtype=float)
    n_samp   = x.shape[0]
    if verbose:
        if f_write is not None:
            f_write.write('\n## optimization paramter:\n')
            f_write.write('# n_itr=%d, n_samp=%d, lambda0=%0.4f, lambda1=%0.4f\n'%(n_itr,n_samp,lambda0,lambda1))           
    # Initialization using pytorch.
    lambda0 = Variable(torch.Tensor([lambda0]),requires_grad=False)
    lambda1 = Variable(torch.Tensor([lambda1]),requires_grad=False)
    p       = Variable(torch.from_numpy(p).float(),requires_grad=False)
    x       = Variable(torch.from_numpy(x).float(),requires_grad=False)
    a       = Variable(torch.Tensor(a),requires_grad=True)
    b       = Variable(torch.Tensor([b]),requires_grad=True)
    w       = Variable(torch.Tensor(w),requires_grad=True)
    mu      = Variable(torch.Tensor(mu),requires_grad=True)
    # Reparametrize by sigma_mean.
    sigma_mean = np.mean(sigma,axis=0)
    sigma      = Variable(torch.Tensor(sigma/sigma_mean),requires_grad=True)
    sigma_mean = Variable(torch.from_numpy(sigma_mean).float(),requires_grad=False)    
    if verbose:
        if f_write is not None:
            f_write.write('\n## method_single_fold: initialization\n')
            f_write.write('# Slope: a=%s\n'%(a.data.numpy()))
            f_write.write('         b=%0.4f\n'%(b.data.numpy()))
            for k in range(K):
                f_write.write('# Bump %d: w=%0.4f\n'%(k,w.data.numpy()[k]))
                f_write.write('         mu=%s\n'%(mu.data.numpy()[k])) 
                f_write.write('      sigma=%s\n'%(sigma.data.numpy()[k])) 
            f_write.write('\n')
    # Specifying the optimizer.
    optimizer = torch.optim.Adam([a,b,w,mu,sigma],lr=0.02)
    scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer,milestones=[100,300,600],gamma=0.5)     
    optimizer.zero_grad()
    # Optimization.
    for l in range(n_itr):        
        scheduler.step()
        # Calculate the model.
        optimizer.zero_grad()
        sigma = sigma.clamp(min=1e-4)
        t = torch.exp(torch.matmul(x,a)+b)
        for i in range(K):
            t = t+torch.exp(w[i]-torch.matmul((x-mu[i,:])**2,sigma[i,:] * sigma_mean))
        loss1 = -torch.mean(torch.sigmoid(lambda0*(t-p)))
        loss2 = lambda1*tf.relu(torch.mean(torch.sigmoid(lambda0*(t+p-1)))\
                                - alpha*torch.mean(torch.sigmoid(lambda0*(t-p))))        
        loss  = loss1+loss2
        # Back propogation
        loss.backward()
        optimizer.step()               
        # Record important quantities.
        loss_rec[l] = loss.data.numpy()        
        if verbose:
            if l%(int(n_itr)/5)==0:
                if f_write is not None:
                    f_write.write('\n## iteration %d: \n'%(l))    
                    f_write.write('# n_rej=%d, n_rej_apr=%0.1f, FD_hat=%d, FD_hat_apr=%0.1f\n'%\
                                  (np.sum(t.data.numpy()>p.data.numpy()),\
                                   np.sum(sigmoid(lambda0.data.numpy()*(t.data.numpy()-p.data.numpy()))),\
                                   np.sum(p.data.numpy()>1-t.data.numpy()),\
                                   np.sum(sigmoid(lambda0.data.numpy()*(p.data.numpy()-1+t.data.numpy())))))
                    f_write.write('# loss1=%0.4f, loss2=%0.4f, FDP=%0.4f, FDP_hat_apr=%0.4f\n'\
                                  %(loss1.data.numpy(),loss2.data.numpy(),\
                                  np.sum((h==0)*(p.data.numpy()<t.data.numpy()))/np.sum(p.data.numpy()<t.data.numpy()),\
                                  (torch.mean(torch.sigmoid(lambda0*(t+p-1)))\
                                  /torch.mean(torch.sigmoid(lambda0*(t-p)))).data.numpy()))
                    f_write.write('# Slope: a=%s\n'%(a.data.numpy()))
                    f_write.write('         b=%0.4f\n'%(b.data.numpy()))
                    for k in range(K):
                        f_write.write('# Bump %d: w=%0.4f\n'%(k,w.data.numpy()[k]))
                        f_write.write('         mu=%s\n'%(mu.data.numpy()[k])) 
                        f_write.write('      sigma=%s\n'%(sigma.data.numpy()[k]))
                if d==1:
                    if output_folder is not None:
                        plt.figure(figsize=[8,5])
                        plot_t(t.data.numpy(),p.data.numpy(),x.data.numpy(),h)
                        plt.savefig(output_folder+'/threshold_itr_%d_fold_%d.png'%(l,fold_number))
                        plt.close()
    if verbose:
        if output_folder is not None:
            plt.figure(figsize=[6,5])
            plt.plot(np.log(loss_rec-loss_rec.min()+1e-3))
            plt.savefig(output_folder+'/loss_fold_%d.png'%fold_number)
            plt.close()      
    # Convert the results.
    p     = p.data.numpy()
    x     = x.data.numpy()
    a     = a.data.numpy()
    b     = b.data.numpy()
    w     = w.data.numpy()
    mu    = mu.data.numpy()
    sigma = (sigma * sigma_mean).data.numpy()
    # Testing.
    t = t_cal(x,a,b,w,mu,sigma)
    gamma = rescale_mirror(t,p,alpha,f_write=f_write,title='after method_single_fold')   
    t *= gamma
    n_rej=np.sum(p<t)     
    if verbose: 
        if f_write is not None:
            f_write.write('\n## Test result with method_single_fold\n')
        result_summary(p<t, h=h, f_write=f_write, title='method_single_fold_%d'%fold_number)
    theta = [a,b,w,mu,sigma,gamma]
    return n_rej, t, theta

def reparametrize(a_init, mu_init, sigma_init, w_init, d):
    """Learn the decision threshold via optimization.

    Args:
        a_init ((d,) ndarray): slope parameter.
        mu_init,sigma_init ((K,d) ndarray): bump parameter.
        w_init ((n,) ndarray): proportion of each component.
        d (int): The dimensionality.     
        
    Returns:
        Parameters used in optimizataion and t_cal.
        a ((d,) ndarray): slope parameter.
        mu,sigma ((K,d) ndarray): bump parameter.
        w ((n,) ndarray): proportion of each component.
    """
    
    a  = a_init
    b  = np.log(w_init[0]).sum()\
        + np.log(a_init[a_init!=0]/(np.exp(a_init[a_init!=0])-1)).sum()    
    K = mu_init.shape[0]
    c = np.zeros([K, d], dtype=float)
    for k in range(K):
        for i in range(d):
            c[k, i] = sp.stats.norm.cdf(1,loc=mu_init[k,i],scale=sigma_init[k,i])\
                - sp.stats.norm.cdf(0,loc=mu_init[k,i],scale=sigma_init[k,i])
    w_init = (w_init + 1e-8)/np.sum(w_init + 1e-8)
    c = c.clip(min=0.1)
    w = np.log(w_init[1:])\
        - d/2*np.log(2*np.pi)\
        - np.log(sigma_init).sum(axis=1)\
        - np.log(c).sum(axis=1)
    mu = mu_init 
    sigma_init = sigma_init.clip(min=1e-3)
    sigma = 1/(2*sigma_init**2)   
    return a,b,w,mu,sigma

def t_cal(x,a,b,w,mu,sigma):
    """Calculate the threshold based on the learned mixture: trend+bump

    Args:
        x ((n,d) ndarray): The covaraites.
        a ((d,) ndarray): reparametrized slope parameter.
        mu,sigma ((K,d) ndarray): reparametrized bump parameter.
        w ((n,) ndarray): reparametrized proportion of each component.
        
    Returns:
        t ((n,) ndarray): threshold for each hypothesis.
    """
    if len(x.shape)==1:
        x = x.reshape([-1, 1])
    t = np.exp(x.dot(a)+b)
    for i in range(mu.shape[0]):
        t += np.exp(w[i])*np.exp(-np.sum((x-mu[i])**2*sigma[i],axis=1))
    return t

def rescale_mirror(t,p,alpha,f_write=None,title=''):
    """Rescale to have the mirror estimate below level alpha

    Args:
        t ((n,) ndarray): The threshold for each hypothesis.
        p ((n,) ndarray): The p-value for each hypothesis.
        alpha (float): The nominal FDR level.
        f_write (file handle): The file to write on.
        title (String): A description of the task. 
        
    Returns:
        (float): the rescale factor.
    """
    if f_write is not None:
        f_write.write('\n## rescale_mirror: %s\n'%title)    
    # Rescale t to a sensible region.
    t_999 = np.percentile(t,99.9)
    if t.clip(max=t_999).mean()>0.2:
        gamma_pre = 0.2/t.clip(max=t_999).mean()
    else: 
        gamma_pre=1
    t = t*gamma_pre
    if f_write is not None:
        f_write.write('# quantile of t (1,25,75,99): %s\n'%(np.percentile(t,[1,25,75,99])))
        f_write.write('# gamma_pre=%0.4f\n'%gamma_pre)
    # Grid search.
    gamma_grid = np.linspace(0, 0.2/np.mean(t), 50)[1:]
    alpha_hat = np.zeros([gamma_grid.shape[0]], dtype=float)
    for i in range(gamma_grid.shape[0]):
        alpha_hat[i] = np.sum(p>1-t*gamma_grid[i])/max(np.sum(p<t*gamma_grid[i]), 1)
    if np.sum(alpha_hat<alpha) > 0:
        gamma_l = np.max(gamma_grid[alpha_hat<alpha])
    else:
        gamma_l = 0
    if np.sum(alpha_hat>alpha) > 0:
        gamma_u = np.min(gamma_grid[alpha_hat>alpha])
    else:
        gamma_u = gamma_grid[-1]
    # Binary search.
    gamma_m = (gamma_u+gamma_l)/2    
    while gamma_u-gamma_l>1e-2 or (np.sum(p>1-t*gamma_m)/max(np.sum(p<t*gamma_m), 1) > alpha):
        gamma_m = (gamma_l+gamma_u)/2
        D_hat = max(np.sum(p<t*gamma_m), 1)
        FD_hat = np.sum(p>1-t*gamma_m)
        alpha_hat = FD_hat/D_hat         
        if f_write is not None:
            f_write.write('# gamma_l=%0.4f, gamma_u=%0.4f, D_hat=%d, FD_hat=%d, alpha_hat=%0.4f\n'%\
                         (gamma_l,gamma_u,D_hat,FD_hat,alpha_hat))
        if alpha_hat < alpha:
            gamma_l = gamma_m
        else: 
            gamma_u = gamma_m    
    if f_write is not None:
        f_write.write('# final output: gamma=%0.4f\n'%((gamma_u+gamma_l)/2*gamma_pre))
    return (gamma_u+gamma_l)/2*gamma_pre 

def method_init(p_input, x_input, K, alpha=0.1, n_full=None, h=None, verbose=False,
                output_folder=None, random_state=0, fold_number=0, f_write=None):
    """Initialization for method_single_fold that fits a mixture model with bump+slope.

    Args:
        p_input ((n,) ndarray): The p-values
        x_input ((n,d) ndarray): The covaraites.
        K (int): The number of bump components.
        alpha (float): The nominal FDR level.
        n_full (int): Total number of hypotheses before filtering.
        h ((n,) ndarray): The ground truth (None if not available).
        verbose (bool): Indicate if output the computation details.
        output_folder (string): The output directory.
        random_state (int): The random seed.
        fold_number (int(0,1)): The fold number.
        f_write (file handler (write mode)): The output file.        
        
    Returns:
        a ((d,) ndarray): slope parameter.
        mu,sigma ((k,d) ndarray): bump parameter.
        w ((n,) ndarray): proportion of each component.
    """
    np.random.seed(random_state)
    p = np.copy(p_input)
    x = np.copy(x_input)
    if f_write is not None:
        f_write.write('## method_init starts\n')        
    if len(x.shape)==1: 
        x = x.reshape([-1,1])
    n_samp,d = x.shape
    if n_full is None:
        n_full = n_samp            
    # Extract the null and the alternative proportion.
    _,t_BH = bh_test(p,n_full=n_full,alpha=0.1)
    if np.sum(p<t_BH) > 100:
        x_null,x_alt = x[p>0.75],x[p<t_BH]
    else:
        x_null,x_alt = x[p>0.75],x[p<0.05]
    if f_write is not None:
        f_write.write('# t_BH=%0.6f, n_null=%d, n_alt=%d\n'%(t_BH, x_null.shape[0],x_alt.shape[0]))      
    # Fit the null distribution.
    if f_write is not None:
        f_write.write('## Learning null distribution\n')            
    a_null,mu_null,sigma_null,w_null = mixture_fit(x_null,K,verbose=verbose,\
                                                   random_state=random_state,f_write=f_write,\
                                                   output_folder=output_folder,\
                                                   suffix='_null',fold_number=fold_number)   
    # Fit the alternative distribution (weighted by null)
    x_w = 1/(f_all(x_alt,a_null,mu_null,sigma_null,w_null)+1e-5)
    x_w /= np.mean(x_w)
    if f_write is not None:
        f_write.write('## Learning alternative distribution\n')  
    a,mu,sigma,w = mixture_fit(x_alt,K,x_w=x_w,verbose=verbose,\
                               random_state=random_state,f_write=f_write,\
                               output_folder=output_folder,\
                               suffix='_alt',fold_number=fold_number)    
    
    if verbose:        
        t = f_all(x,a,mu,sigma,w)
        gamma = rescale_mirror(t,p,alpha)   
        t = t*gamma
        if f_write is not None:
            f_write.write('\n## Test result with method_init\n')        
        result_summary(p<t,h,f_write=f_write, title='method_init_%d'%fold_number)  
        # Plot the fitted results
        if output_folder is not None:
            if d==1:
                plt.figure(figsize=[8,5])
            else:
                plt.figure(figsize=[8,12])
            plot_t(t,p,x,h)        
            plt.tight_layout()
            plt.savefig(output_folder+'/threshold_after_method_init_fold_%d.png'%fold_number)
        else:
            plt.show()
        if f_write is not None:
            f_write.write('## method_init finished\n')
    return a,mu,sigma,w    

def mixture_fit(x,K=3,x_w=None,n_itr=100,verbose=False,random_state=0,f_write=None,\
                output_folder=None,suffix=None,fold_number=0): 
    """Fit a slope+bump mixture using EM algorithm.

    Args:
        x ((n,d) ndarray): The covaraites.
        K (int): The number of bump components.
        x_w ((n,) ndarray): The weights for each sample.
        n_itr (int): The maximum number of iterations for the EM algorithm
        verbose (bool): Indicate if output the computation details.
        random_state (int): The random seed.
        f_write (file handler (write mode)): The output file.
        output_folder (string): The output directory.
        suffix (string): The suffix of the output file.
        fold_number (int(0,1)): The fold number.

    Returns:
        a ((d,) ndarray): slope parameter.
        mu,sigma ((k,d) ndarray): bump parameter.
        w ((n,) ndarray): proportion of each component. 
    """
 
    np.random.seed(random_state)
    if len(x.shape)==1: 
        x = x.reshape([-1,1])
    n_samp,d = x.shape
    if x_w is None: 
        x_w=np.ones([n_samp],dtype=float)        
    # Initialization
    GMM      = GaussianMixture(n_components=K,covariance_type='diag').fit(x)
    w_old    = np.zeros([K+1])
    w        = 0.5*np.ones([K+1])/K
    w[0]     = 0.5
    a        = ML_slope(x,x_w)   
    mu,sigma = GMM.means_,GMM.covariances_**0.5
    w_samp   = np.zeros([K+1,n_samp],dtype=float)
    i        = 0
    # Print the initialization information
    if verbose:
        if f_write is not None:
            f_write.write('## mixture_fit: initialization parameters\n')
            f_write.write('# Slope: w=%0.4f, a=%s\n'%(w[0],a))
            for k in range(K):
                f_write.write('# Bump %d: w=%0.4f\n'%(k,w[k+1]))
                f_write.write('         mu=%s\n'%(mu[k])) 
                f_write.write('      sigma=%s\n'%(sigma[k])) 
            f_write.write('\n')
    # EM algorithm        
    while np.linalg.norm(w-w_old,1)>5e-3 and i<n_itr:              
        # E step       
        w_old = w
        w_samp[0,:] = w[0]*f_slope(x,a)
        for k in range(K):
            w_samp[k+1,:] = w[k+1]*f_bump(x,mu[k],sigma[k])
        w_samp = w_samp/np.sum(w_samp,axis=0)*x_w             
        # M step
        w = np.mean(w_samp,axis=1) 
        a = ML_slope(x,w_samp[0,:])
        for k in range(K):
            if w[k+1]>1e-4: mu[k],sigma[k]=ML_bump(x,w_samp[k+1,:])               
        sigma = sigma.clip(min=1e-4)
        w[w<1e-3] = 0
        w /= w.sum()
        i += 1        
    # Convergence warning       
    if i >= n_itr and verbose: 
        print('Warning: the model does not converge, w_dif=%0.4f'%np.linalg.norm(w-w_old,1))        
        if f_write is not None:
            f_write.write('Warning: the model does not converge, w_dif=%0.4f\n'%np.linalg.norm(w-w_old,1))
    # Output 
    if verbose and f_write is not None:
        f_write.write('## mixture_fit: learned parameters\n')
        f_write.write('# Slope: w=%0.4f, a=%s\n'%(w[0],a))
        for k in range(K):
            f_write.write('# Bump %d: w=%0.4f\n'%(k,w[k+1]))
            f_write.write('         mu=%s\n'%(mu[k])) 
            f_write.write('      sigma=%s\n'%(sigma[k])) 
        f_write.write('\n')        
    if output_folder is not None:
        bins_ = np.linspace(0,1,101)
        x_grid = bins_.reshape([-1,1])       
        if d==1:
            plt.figure(figsize=[8,5])
            plt.hist(x,bins=bins_,weights=x_w/np.sum(x_w)*100) 
            temp_p = f_all(x_grid,a,mu,sigma,w)      
            plt.plot(bins_,temp_p)
            plt.savefig(output_folder+'/projection%s_fold_%d.png'%(suffix,fold_number))        
        else:
            plt.figure(figsize=[8,12])
            n_figure = min(d, 5)
            for i_dim in range(n_figure):        
                plt.subplot(str(n_figure)+'1'+str(i_dim+1))
                plt.hist(x[:,i_dim],bins=bins_,weights=x_w/np.sum(x_w)*100)  
                temp_p = f_all(x_grid,a[[i_dim]],mu[:,[i_dim]],sigma[:,[i_dim]],w)  
                plt.plot(bins_,temp_p)
                plt.title('Dimension %d'%(i_dim+1))
            plt.savefig(output_folder+'/projection%s_fold_%d.png'%(suffix,fold_number))
        plt.close('all')    
    return a,mu,sigma,w

"""
    sub-routines for mixture_fit
    input:  x: (n,d) ndarray (not (n,))
            w: (n,) ndarray
            a: (d,) ndarray
            mu,sigma: (k,d) ndarray
"""
def ML_slope(x,v=None,c=0.005):  
    """
    ML fit of the slope a/(e^a-1) e^(ax), defined over [0,1]

    Parameters
    ----------
    x : (n,d) ndarray
        covaraites
    v : (n,) ndarray 
        weight for each sample
    c : float
        regularization factor 

    Returns
    -------
    (d,) ndarray
        the estiamted slope parameter
    """    
    n,d = x.shape
    a = np.zeros(d,dtype=float)
    if v is None:
        t = np.mean(x,axis=0) 
    else:
        t = np.sum((x.T*v).T,axis=0)/np.sum(v) # weighted sum along each dimension
    
    a_u=10*np.ones([d],dtype=float)
    a_l=-10*np.ones([d],dtype=float)
    # binary search 
    while np.linalg.norm(a_u-a_l)>0.01:      
        a_m  = (a_u+a_l)/2
        a_m += 1e-2*(a_m==0)
        temp = (np.exp(a_m)/(np.exp(a_m)-1) - 1/a_m + 2*c*a_m)
        a_l[temp<t] = a_m[temp<t]
        a_u[temp>=t] = a_m[temp>=t]
    return (a_u+a_l)/2
    
def f_slope(x,a):
    """
    density of the slope function

    Parameters
    ----------
    x : (n,d) ndarray
        coML estimatearaites
    a : (d,) ndarray 
        slope parameter (for each dimension)

    Returns
    -------
    (n,) array
        the slope density for each point
    """
    f_x = np.exp(x.dot(a)) # dimension-wise probability
    norm_factor = np.prod(a[a!=0]/(np.exp(a[a!=0])-1))
    f_x = f_x * norm_factor # actual probability     

    return f_x

## Very hard to vectorize. So just leave it here
def ML_bump(x,v=None,logger=None):
    """
    ML fit of the bump function

    Parameters
    ----------
    x : (n,d) ndarray
        coML estimatearaites
    v : (n,) ndarray 
        weight for each sample

    Returns
    -------
    mu : (n,d) ndarray
        bump mean parameter (for each dimension)
    sigma : (n,d) ndarray 
        bump std parameter (for each dimension)
    """   
    def ML_bump_1d(x,v,logger=None):
        def fit_f(param,x,v):                
            mu,sigma = param
            inv_sigma = 1/sigma
            Z = sp.stats.norm.cdf(1,loc=mu,scale=sigma)-sp.stats.norm.cdf(0,loc=mu,scale=sigma)            
            inv_Z = 1/Z
            phi_alpha = 1/np.sqrt(2*np.pi)*np.exp(-mu**2/2/sigma**2)
            phi_beta = 1/np.sqrt(2*np.pi)*np.exp(-(1-mu)**2/2/sigma**2)

            # Average likelihood
            if v is None:
                t1 = np.mean(x-mu)
                t2 = np.mean((x-mu)**2)
            else:
                t1 = np.sum((x-mu)*v) / np.sum(v)
                t2 = np.sum((x-mu)**2*v) / np.sum(v)
                
            l = -np.log(Z) - np.log(sigma) - t2/2/sigma**2        
            # Gradient    
            d_c_mu = inv_sigma * (phi_alpha-phi_beta)
            d_c_sig = inv_sigma * (-mu*inv_sigma*phi_alpha - (1-mu)*inv_sigma*phi_beta)
            d_l_mu = -d_c_mu*inv_Z + t1*inv_sigma**2
            d_l_sig = -d_c_sig*inv_Z - inv_sigma + t2*inv_sigma**3
            grad = np.array([d_l_mu,d_l_sig],dtype=float)  
            return l,grad
        
        ## gradient check
        #_,grad_ = fit_f([0.2,0.1],x,v)
        #num_dmu = (fit_f([0.2+1e-8,0.1],x,v)[0]-fit_f([0.2,0.1],x,v)[0]) / 1e-8
        #num_dsigma = (fit_f([0.2,0.1+1e-8],x,v)[0]-fit_f([0.2,0.1],x,v)[0]) / 1e-8          
        #print('## Gradient check ##')
        #print('#  param value: mu=%0.6f, sigma=%0.6f'%(0.2,0.1))
        #print('#  Theoretical grad: dmu=%0.8f, dsigma=%0.8f'%(grad_[0],grad_[1]))
        #print('#  Numerical grad: dmu=%0.8f, dsigma=%0.8f\n'%(num_dmu,num_dsigma))
            
        # If the variance is small and the mean is at center, 
        # directly output the empirical mean and variance.
        if v is None:
            mu = np.mean(x)
            sigma = np.std(x)
        else:       
            mu = np.sum(x*v)/np.sum(v)
            sigma = np.sqrt(np.sum((x-mu)**2*v)/np.sum(v))
        
        if sigma<0.075 and np.min([1-mu,mu])>0.15:
            return mu,sigma
        
        param = np.array([mu,sigma])    
        lr = 0.01
        max_step = 0.025
        max_itr = 100
        i_itr = 0
        l_old = -10
        
        while i_itr<max_itr:            
            l,grad = fit_f(param,x,v)
            if np.absolute(l-l_old)<0.001:
                break
            else:
                l_old=l
            update = (grad*lr).clip(min=-max_step,max=max_step)
            param += update
            i_itr +=1                 
            if np.isnan(param).any() or np.min([param[0],1-param[0],param[1]])<0:  
                return np.mean(x),np.std(x)
                
        mu,sigma = param  
        if sigma>0.25:
            sigma=1
        return mu,sigma      
  
    mu    = np.zeros(x.shape[1],dtype=float)
    sigma = np.zeros(x.shape[1],dtype=float)
    for i in range(x.shape[1]):
        mu[i],sigma[i] = ML_bump_1d(x[:,i],v,logger=logger)
    return mu,sigma

def f_bump(x,mu,sigma):
    """
    density of the bump function

    Parameters
    ----------
    x : (n,d) ndarray
        coML estimatearaites
    mu : (d,) ndarray 
        bump mean parameter (for each dimension)
     
    sigma : (d,) ndarray 
        bump std parameter (for each dimension)

    Returns
    -------
    (n,) array
        the bump density for each point
    """    
    def f_bump_1d(x,mu,sigma):
        if sigma<1e-6: 
            return np.zeros(x.shape[0],dtype=float)
        inv_sigma=1/sigma
        pmf = sp.stats.norm.cdf(1,loc=mu,scale=sigma)-sp.stats.norm.cdf(0,loc=mu,scale=sigma)
        return inv_sigma/np.sqrt(2*np.pi)*np.exp(-inv_sigma**2*(x-mu)**2/2)/pmf
       
    f_x = np.ones([x.shape[0]],dtype=float)
    for i in range(x.shape[1]):
        f_x *=  f_bump_1d(x[:,i],mu[i],sigma[i])
    return f_x       
        
def f_all(x,a,mu,sigma,w):
    """
    density of the mixture model (slope+bump)

    Parameters
    ----------
    x : (n,d) ndarray
        coML estimatearaites
    mu : (d,) ndarray 
        bump mean parameter (for each dimension)
     
    sigma : (d,) ndarray 
        bump std parameter (for each dimension)

    Returns
    -------
    (n,) array
        the bump density for each point
    """
    
    f = w[0]*f_slope(x,a)        
    for k in range(1,w.shape[0]):
        f += w[k]*f_bump(x,mu[k-1],sigma[k-1])           
    return f

'''
    Baseline comparison methods
'''
def bh_test(p,alpha=0.1,n_full=None,verbose=False):
    if n_full is None: 
        n_full = p.shape[0]
    p_sort   = sorted(p)
    n_rej    = 0
    for i in range(p.shape[0]):
        if p_sort[i] < i*alpha/n_full:
            n_rej = i
    t_rej = p_sort[n_rej]
    if verbose:
        print("## bh testing summary ##")
        print("# n_rej = %d"%n_rej)
        print("# t_rej = %0.6f"%t_rej)
        print("\n")
    return n_rej,t_rej

def sbh_test(p,alpha=0.1,lamb=0.5,n_full=None,verbose=False):
    if n_full is None: 
        n_full = p.shape[0]
    if n_full > p.shape[0]:
        lamb = np.min(p[p>0.5])
    print('lamb= %0.4f'%lamb)
    pi0_hat  = (np.sum(p>lamb)/(1-lamb)/n_full).clip(max=1)  
    alpha   /= pi0_hat
    print('## pi0_hat=%0.3f'%pi0_hat) 
    p_sort = sorted(p)
    n_rej = 0
    for i in range(p.shape[0]):
        if p_sort[i] < i*alpha/n_full:
            n_rej = i
    t_rej = p_sort[n_rej]
    if verbose:
        print("## sbh summary ##")
        print("# n_rej = %d"%n_rej)
        print("# t_rej = %0.6f"%t_rej)
        print("# pi_0 estimate = %0.3f"%pi0_hat)
        print("\n")
    return n_rej,t_rej,pi0_hat

"""
    some old code
""" 


## old code for feature_explore
#def feature_explore(p,x_,alpha=0.1,qt_norm=False,vis_dim=None,cate_name={},output_folder=None,h=None):
#    def plot_feature_1d(x_margin,p,x_null,x_alt,bins,meta_info,title='',cate_name=None,\
#                        output_folder=None,h=None):
#        feature_type,cate_order = meta_info        
#        if feature_type == 'continuous':         
#            ## continuous feature: using kde to estimate 
#            n_bin = bins.shape[0]-1
#            x_grid = (bins+(bins[1]-bins[0])/2)[0:-1]
#            p_null,_ = np.histogram(x_null,bins=bins) 
#            p_alt,_= np.histogram(x_alt,bins=bins)         
#            p_null = p_null/np.sum(p_null)*n_bin
#            p_alt = p_alt/np.sum(p_alt)*n_bin
#            kde_null = stats.gaussian_kde(x_null).evaluate(x_grid)
#            kde_alt = stats.gaussian_kde(x_alt).evaluate(x_grid)
#            p_ratio = (kde_alt+1e-2)/(kde_null+1e-2)        
#                 
#        else: 
#            ## discrete feature: directly use the empirical counts 
#            unique_null,cts_null = np.unique(x_null,return_counts=True)
#            unique_alt,cts_alt = np.unique(x_alt,return_counts=True)            
#            unique_val = np.array(list(set(list(unique_null)+list(unique_alt))))
#            unique_val = np.sort(unique_val)            
#            p_null,p_alt = np.zeros([unique_val.shape[0]]),np.zeros([unique_val.shape[0]])          
#            for i,key in enumerate(unique_null): 
#                p_null[unique_val==key] = cts_null[i]                
#            for i,key in enumerate(unique_alt): 
#                p_alt[unique_val==key] = cts_alt[i]           
#            n_bin = unique_val.shape[0]           
#            p_null = (p_null+1)/np.sum(p_null+1)*n_bin
#            p_alt = (p_alt+1)/np.sum(p_alt+1)*n_bin            
#            p_ratio = (p_alt+1e-2)/(p_null+1e-2)  
#            x_grid = (np.arange(unique_val.shape[0])+1)/(unique_val.shape[0]+1)
#            
#            if cate_name is None: 
#                cate_name_ = cate_order
#            else:
#                cate_name_ = []
#                for i in cate_order:
#                    cate_name_.append(cate_name[i])
#                    
#        plt.figure(figsize=[8,8])
#        plt.subplot(311)
#        rnd_idx=np.random.permutation(p.shape[0])[0:np.min([10000,p.shape[0]])]
#        p = p[rnd_idx]
#        x_margin = x_margin[rnd_idx]
#        
#        if h is not None:
#            plt.scatter(x_margin[h==1],p[h==1],color='orange',alpha=0.3,s=4,label='alt')
#            plt.scatter(x_margin[h==0],p[h==0],color='royalblue',alpha=0.3,s=4,label='null')
#        else:
#            plt.scatter(x_margin,p,color='royalblue',alpha=0.3,s=4,label='alt')
#        plt.ylim([0,(p[p<0.5].max())])
#        plt.ylabel('p-value')
#        plt.title(title+' (%s)'%feature_type)
#        plt.subplot(312)
#        plt.bar(x_grid,p_null,width=1/n_bin,color='royalblue',alpha=0.6,label='null')
#        plt.bar(x_grid,p_alt,width=1/n_bin,color='darkorange',alpha=0.6,label='alt')
#        plt.xlim([0,1])
#        if feature_type=='discrete': 
#            plt.xticks(x_grid,cate_name_,rotation=45)
#        plt.ylabel('null/alt proportion')
#        plt.legend()
#        plt.subplot(313)
#        if feature_type == 'continuous':
#            plt.plot(x_grid,p_ratio,color='seagreen',label='ratio',linewidth=4) 
#        else:
#            plt.plot(x_grid,p_ratio,color='seagreen',marker='o',label='ratio',linewidth=4)
#            plt.xticks(x_grid,cate_name_,rotation=45)
#        plt.xlim([0,1])
#        plt.ylabel('ratio')
#        plt.xlabel('Covariate $x$')
#        plt.tight_layout()
#        if output_folder is not None:
#            plt.savefig(output_folder+'/explore_%s.png'%title)
#        plt.show()
#    
#    
#    ## preprocessing
#    x,meta_info = feature_preprocess(x_,p,qt_norm=qt_norm,continue_rank=False,require_meta_info=True)   
#    x_nq,meta_info = feature_preprocess(x_,p,qt_norm=False,continue_rank=False,require_meta_info=True)   
#    d = x.shape[1]
#    
#    ## separate the null proportion and the alternative proportion
#    _,t_BH = bh(p,alpha=alpha)
#    x_null,x_alt = x[p>0.5],x[p<t_BH]      
#    x_null_nq,x_alt_nq = x_nq[p>0.5],x_nq[p<t_BH]      
#    
#    ## generate the figure
#    bins = np.linspace(0,1,26)  
#    if vis_dim is None: 
#        vis_dim = np.arange(min(4,d))    
#    
#    for i in vis_dim:
#        x_margin = x[:,i]
#        if meta_info[i][0] == 'continuous':
#            temp_null,temp_alt = x_null[:,i],x_alt[:,i]
#        else:
#            temp_null,temp_alt = x_null_nq[:,i],x_alt_nq[:,i]
#        if i in cate_name.keys():
#            plot_feature_1d(x_margin,p,temp_null,temp_alt,bins,meta_info[i],title='feature_%s'%str(i+1),\
#                            cate_name=cate_name[i],output_folder=output_folder,h=h)
#        else:
#            plot_feature_1d(x_margin,p,temp_null,temp_alt,bins,meta_info[i],title='feature_%s'%str(i+1),\
#                            output_folder=output_folder,h=h)
#    return

    #def ML_slope_1d(x,w=None,c=0.05):
    #    if w is None:
    #        w = np.ones(x.shape[0])        
    #    t = np.sum(w*x)/np.sum(w) ## sufficient statistic
    #    a_u=100
    #    a_l=-100
    #    ## binary search 
    #    while a_u-a_l>0.1:
    #        a_m  = (a_u+a_l)/2
    #        a_m += 1e-2*(a_m==0)
    #        if (np.exp(a_m)/(np.exp(a_m)-1) - 1/a_m +c*a_m)<t:
    #            a_l = a_m
    #        else: 
    #            a_u = a_m
    #    return (a_u+a_l)/2
    #
    #a = np.zeros(x.shape[1],dtype=float)
    #for i in range(x.shape[1]):
    #    a[i] = ML_slope_1d(x[:,i],w,c=c)
    #return a
    


"""

This module contains all functions that are used the load the data.

Todo:
    * Clean the code.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html
   
Format for data loaders:
    p, x, h, n_full, cate_name

"""
import numpy as np 
import scipy as sp
import pickle
from scipy import stats
from scipy.stats import ttest_ind
import matplotlib.pyplot as plt
import adafdr
from adafdr.util import *
from matplotlib import mlab
from adafdr.method import *
import logging

# External datasets
def data_airway():
    file_path = adafdr.__path__[0]
    file_name = file_path + '/data/airway'
    X = np.loadtxt(file_name,skiprows=0,delimiter=',')
    x=X[:,2].reshape([-1,1])
    p=X[:,0]
    return p, x

def data_bottomly():
    file_path = adafdr.__path__[0]
    file_name = file_path + '/data/bottomly'
    X = np.loadtxt(file_name,skiprows=0,delimiter=',')
    x=X[:,2].reshape([-1,1])
    p=X[:,0]
    return p, x

def data_pasilla():
    file_path = adafdr.__path__[0]
    file_name = file_path + '/data/pasilla'
    X = np.loadtxt(file_name,skiprows=0,delimiter=',')
    x=X[:,2].reshape([-1,1])
    p=X[:,0]
    return p, x

def data_small_gtex():
    # Hard-coded information of the GTEx dataset.
    cate_name = {3: {1: 'TssA', 2: 'TssAFlnk', 3: 'TxFlnk', 4: 'Tx',
                     5: 'TxWk', 6: 'EnhG', 7: 'Enh', 8: 'ZNF/Rpts',
                     9: 'Het', 10: 'TssBiv', 11: 'BivFlnk', 12: 'EnhBiv',
                     13: 'ReprPC', 14: 'ReprPCWk', 15: 'Quies'}}
    n_full = 172353475
    fname = 'GTEx_small.pickle'
    file_path = adafdr.__path__[0]
    fname = file_path + '/data/' + fname
    with open(fname, 'rb') as handle:  
        p = pickle.load(handle)
        x = pickle.load(handle)
        cis_name = pickle.load(handle)
    return p, x, n_full, cate_name, cis_name

## generating the 1d toy example
def toy_data_1d(job_id=0,n_sample=10000,vis=0):
    def pi1_gen(x): # need to be fixed 
        pi1=0.03*sp.stats.norm.pdf(x,loc=0.2,scale=0.05)+0.04*sp.stats.norm.pdf(x,loc=0.8,scale=0.05)
        pi1+=0.15*x 
        return pi1

    def plot_pi1_1d(pi1_gen):
        x_grid   = np.linspace(0,1,100)
        pi1_grid = pi1_gen(x_grid)  
        plt.plot(x_grid,pi1_grid)
        plt.xlabel('covariate')
        plt.ylabel('alt distribution')
        plt.title('the alternative distribution')
    
    np.random.seed(42)
    if job_id == 0: # Gaussian mixtures     
        x   = np.random.uniform(0,1,size=n_sample)
        pi1 = pi1_gen(x)
        p   = np.zeros(n_sample)
        
        # generating the hypothesis       
        h  = np.array((np.random.uniform(size=n_sample)<pi1),dtype=int)
        n0 = np.sum(h==0)
        n1 = np.sum(h==1)
        
        # generating the p-values 
        p[h==0] = np.random.uniform(size=n0)
        p[h==1] = np.random.beta(a=0.4,b=4,size=n1)
        
        #plt.figure()
        #plt.hist(p[h==1],bins=100)
        #plt.show()
        #print(np.mean(p[h==1]))
        
        if vis == 1:
            print('### Summary ###')
            print('# null: %s, # alt: %s:, null proportion: %s'%(str(np.sum(h==0)),str(np.sum(h==1)),str(np.sum(h==0)/h.shape[0])))
            plt.figure(figsize=[16,5])
            plt.subplot(121)
            plot_pi1_1d(pi1_gen)
            plt.subplot(122)
            plot_data_1d(p,x,h)
            plt.legend()
            plt.show() 
        return p,x,h
    
def write_simulation_data(p, x, h, filename):
    """Write the simulation data with format:
    p, h, x0, x1, x2, ... for the columns

    Args:
        p ((n,) ndarray): The p-value.
        x ((n,d) ndarray): The covaraites.
        h ((n,) boolean ndarray): The ground truth. True indicates the 
            hypothesis is alternative. 
        filename (str): path of the file.
        
    Returns:
    """
    temp_data = np.zeros([x.shape[0], x.shape[1]+2], dtype=float)
    temp_data[:, 0] = p
    temp_data[:, 1] = h
    temp_data[:, 2:] = x
    np.savetxt(filename, temp_data, delimiter=",")
    return 

def load_simulation_data(filename):
    """Load the simulation data with format:
    p, h, x0, x1, x2, ... for the columns

    Args:
        filename (str): path of the file.
        
    Returns:
        p ((n,) ndarray): The p-value.
        x ((n,d) ndarray): The covaraites.
        h ((n,) boolean ndarray): The ground truth. True indicates the 
            hypothesis is alternative. 
    """
    temp_data = np.loadtxt(filename, delimiter=',')
    p = temp_data[:, 0].astype(float)
    h = temp_data[:, 1].astype(bool)
    x = temp_data[:, 2:].astype(float)
    return p, x, h

def load_x_mixture(opt=0):
    """Generate a mixture data (of x) to test mixture_fit.

    Args:
        opt (int): 0: 2d slope. 
            1: 2d bump. 
            2: 2d slope+bump.
            3: 10d data with slope+bump in the first 2d.

    Returns:
        x ((n,d) ndarray): The mixture data
        param (list): Parameters that are used to generate the data.    

    """
    n_sample = 10000    
    if opt==0:
        a = np.array([2,0],dtype=float)
        x_grid = get_grid_2d(101)
        n_grid = x_grid.shape[0]
        p    = f_slope(x_grid,a)
        p   /= p.sum()
        x = np.random.choice(np.arange(n_grid),size=n_sample,p=p)
        x = x_grid[x,:]
        param = a        
    elif opt==1:  
        mu = np.array([0.5,0.05],dtype=float)
        sigma = np.array([0.1,0.1],dtype=float)
        x_grid = get_grid_2d(101)
        n_grid = x_grid.shape[0]
        p    = f_bump(x_grid,mu,sigma)
        p   /= p.sum()
        x = np.random.choice(np.arange(n_grid),size=n_sample,p=p)
        x = x_grid[x,:]
        param = (mu,sigma)
    elif opt==2:
        w = np.array([0.4,0.3,0.3],dtype=float)
        a = np.array([2,0],dtype=float)
        mu = np.array([[0.2,0.2],[0.7,0.7]],dtype=float)
        sigma = np.array([[0.1,0.2],[0.1,0.1]],dtype=float)
        
        x_grid = get_grid_2d(101)
        n_grid = x_grid.shape[0]
        p = f_all(x_grid,a,mu,sigma,w)
        p   /= p.sum()
        x = np.random.choice(np.arange(n_grid),size=n_sample,p=p)
        x = x_grid[x,:]
        param = (a,mu,sigma,w)            
    elif opt==3:
        w = np.array([0.4,0.3,0.3],dtype=float)
        a = np.array([2,0],dtype=float)
        mu = np.array([[0.2,0.2],[0.7,0.7]],dtype=float)
        sigma = np.array([[0.1,0.2],[0.1,0.1]],dtype=float)
        
        x_grid = get_grid_2d(101)
        n_grid = x_grid.shape[0]
        p = f_all(x_grid,a,mu,sigma,w)
        p   /= p.sum()
        x = np.random.choice(np.arange(n_grid),size=n_sample,p=p)
        x = x_grid[x,:]
        
        a_ = np.zeros(10)
        a_[0:2] = a
        mu_ = np.zeros([2,10],dtype=float)+0.5
        mu_[:,0:2] = mu
        sigma_ = np.ones([2,10],dtype=float)
        sigma_[:,0:2] = sigma
        param = (a_,mu_,sigma_,w)        
        x_noise = np.random.uniform(high=1,low=0,size = (n_sample,8))
        x = np.concatenate([x,x_noise],1)
    else:
        pass    
    return x,param

def load_1d_bump_slope(n_sample=20000, n_dim=2, random_state=0):
    """Generate a 1d simulated data.

    Args:
        n_sample (int): The number of hypotheses.
        n_dim (int): The number of dimensions. If n_dim>2, the rest of dimensions 
            contains uninformative features.
        random_state (int): The random seed

    Returns:
        p ((n,) ndarray): The p-value.
        x ((n,d) ndarray): The covaraites.
        h ((n,) boolean ndarray): The ground truth. True indicates the 
            hypothesis is alternative. 
        n_full (int): The number of hypotheses before filtering. Same as 
            n if no filtering is applied.
        cate_name (dic of dics): (key,val) gives the (feature index, cate_name_dic) for
            discrete features. For each discrete feature, the (key,val) of the sub dic
            gives the (val,name) for all categories.
    """
    np.random.seed(random_state)
    # Generate pi1
    x_grid = get_grid_1d(101)
    x = np.random.choice(np.arange(x_grid.shape[0]), size=n_sample)
    x = x_grid[x,:]       
    w = np.array([0.5,0.25,0.25],dtype=float)
    a = np.array([0.5],dtype=float)
    mu = np.array([[0.25], [0.75]],dtype=float)
    sigma = np.array([[0.05], [0.05]],dtype=float)   
    pi1 = (0.1*f_all(x,a,mu,sigma,w)).clip(max=1)
    # Generate data
    p = np.zeros(n_sample)
    h = np.zeros(n_sample, dtype=bool)
    rnd = np.random.uniform(size=n_sample)    
    p[rnd>=pi1] = np.random.uniform(size=np.sum(rnd>=pi1))
    p[rnd<pi1] = np.random.beta(a=0.3, b=4, size=np.sum(rnd<pi1))
    h[rnd<pi1] = True
    # Add non-informative dimensions.
    if n_dim>1:
        x_noise = np.random.uniform(size=(n_sample, n_dim-2))
        x = np.concatenate([x,x_noise],1)
    return p,x,h,p.shape[0],{}

def load_2d_bump_slope(n_sample=20000, n_dim=2, random_state=0):
    """Generate a simulated data.

    Args:
        n_sample (int): The number of hypotheses.
        n_dim (int): The number of dimensions. If n_dim>2, the rest of dimensions 
            contains uninformative features.
        random_state (int): The random seed

    Returns:
        p ((n,) ndarray): The p-value.
        x ((n,d) ndarray): The covaraites.
        h ((n,) boolean ndarray): The ground truth. True indicates the 
            hypothesis is alternative. 
        n_full (int): The number of hypotheses before filtering. Same as 
            n if no filtering is applied.
        cate_name (dic of dics): (key,val) gives the (feature index, cate_name_dic) for
            discrete features. For each discrete feature, the (key,val) of the sub dic
            gives the (val,name) for all categories.
    """
    np.random.seed(random_state)
    # Generate pi1
    x_grid = get_grid_2d(101)
    x = np.random.choice(np.arange(x_grid.shape[0]),size=n_sample)
    x = x_grid[x,:]       
    w = np.array([0.5,0.25,0.25],dtype=float)
    a = np.array([0.5,0.5],dtype=float)
    mu = np.array([[0.25,0.25],[0.75,0.75]],dtype=float)
    sigma = np.array([[0.1,0.1],[0.1,0.1]],dtype=float)   
    pi1 = (0.1*f_all(x,a,mu,sigma,w)).clip(max=1)
    # Generate data
    p = np.zeros(n_sample)
    h = np.zeros(n_sample, dtype=bool)
    rnd = np.random.uniform(size=n_sample)    
    p[rnd>=pi1] = np.random.uniform(size=np.sum(rnd>=pi1))
    p[rnd<pi1] = np.random.beta(a=0.3, b=4, size=np.sum(rnd<pi1))
    h[rnd<pi1] = True
    # Add non-informative dimensions.
    if n_dim>2:
        x_noise = np.random.uniform(size=(n_sample, n_dim-2))
        x = np.concatenate([x,x_noise],1)
    return p,x,h,p.shape[0],{}

def load_data_ihw(random_state=0):
    """data from ihw supp 4.2.2
    """
    np.random.seed(random_state)
    n_sample = 20000
    n_alt = int(20000*0.1)
    h = np.zeros([n_sample], dtype=int)
    h[0:n_alt] = 1
    data_case = np.random.randn(5, n_sample) + h*2
    data_control = np.random.randn(5, n_sample)
    p = ttest_ind(data_case, data_control)[1]
    data_pool = np.concatenate([data_case, data_control], axis=0)
    x = np.var(data_pool, axis=0)
    x = x.reshape([-1,1])
    return p, x, h

def load_data_wd(n_sample=20000, random_state=0):
    """Weakly dependent dataset following the receipe of Sec. 3.2, 
    from the paper "Strong control, conservative point estimation
    and simultaneous conservative consistency of false discovery rates:
    a unified approach"
    
    """
    np.random.seed(random_state)
    # Weakly-dependent covariance matrix.
    cov_mat = np.zeros([10, 10], dtype=float)
    for j in range(10):
        for k in range(j,10):
            if j == k:
                cov_mat[j, k] = 1
            elif j < k and k <= 4:
                cov_mat[j, k] = 0.25
            elif j <= 4 and k > 4:
                cov_mat[j, k] = -0.25
    for j in range(10):
            for k in range(j):
                cov_mat[j, k] = cov_mat[k, j]
    # Generate pi1
    x_grid = get_grid_1d(101)
    x = np.random.choice(np.arange(x_grid.shape[0]), size=n_sample)
    x = x_grid[x,:]       
    w = np.array([0.5,0.25,0.25],dtype=float)
    a = np.array([0.5],dtype=float)
    mu = np.array([[0.25], [0.75]],dtype=float)
    sigma = np.array([[0.05], [0.05]],dtype=float)   
    pi1 = (0.1*f_all(x,a,mu,sigma,w)).clip(max=1)
    # Generate p-values
    null_z = np.random.multivariate_normal(np.zeros([10]), cov_mat, int(n_sample/10)).flatten()
    alt_z = np.random.multivariate_normal(np.zeros([10])+2, cov_mat, int(n_sample/10)).flatten()
    null_p = 1 - stats.norm.cdf(null_z)
    alt_p = 1 - stats.norm.cdf(alt_z)
    rnd = np.random.uniform(size=n_sample)   
    p = np.zeros([n_sample], dtype=float)
    p[rnd>=pi1] = null_p[0:np.sum(rnd>=pi1)]
    p[rnd<pi1] = alt_p[0:np.sum(rnd<pi1)]
    h = np.zeros(n_sample, dtype=bool)
    h[rnd<pi1] = True
    return p, x, h

def load_data_sd(n_sample=20000, random_state=0):
    """Strong dependent dataset modelling LD: every 5 hypotheses 
    are completely dependent
    
    """
    np.random.seed(random_state)
    # Strongly-dependent covariance matrix.
    cov_mat = np.ones([5, 5], dtype=float)
    # Generate pi1
    x_grid = get_grid_1d(101)
    x = np.random.choice(np.arange(x_grid.shape[0]), size=n_sample)
    x = x_grid[x,:]       
    w = np.array([0.5,0.25,0.25],dtype=float)
    a = np.array([0.5],dtype=float)
    mu = np.array([[0.25], [0.75]],dtype=float)
    sigma = np.array([[0.05], [0.05]],dtype=float)   
    pi1 = (0.1*f_all(x,a,mu,sigma,w)).clip(max=1)
    # Generate p-values
    null_z = np.random.multivariate_normal(np.zeros([5]), cov_mat, int(n_sample/5)).flatten()
    alt_z = np.random.multivariate_normal(np.zeros([5])+2, cov_mat, int(n_sample/5)).flatten()
    null_p = 1 - stats.norm.cdf(null_z)
    alt_p = 1 - stats.norm.cdf(alt_z)
    rnd = np.random.uniform(size=n_sample)   
    p = np.zeros([n_sample], dtype=float)
    p[rnd>=pi1] = null_p[0:np.sum(rnd>=pi1)]
    p[rnd<pi1] = alt_p[0:np.sum(rnd<pi1)]
    h = np.zeros(n_sample, dtype=bool)
    h[rnd<pi1] = True
    return p, x, h

## neuralFDR simulated examples
def neuralfdr_generate_data_1D(job=0, n_samples=10000,data_vis=0, num_case=4):
    if job == 0: # discrete case
        pi1=np.random.uniform(0,0.3,size=num_case)
        X=np.random.randint(0, num_case, n_samples)
        
        p = np.zeros(n_samples)
        h = np.zeros(n_samples)
        
        for i in range(n_samples):
            rnd = np.random.uniform()
            if rnd > pi1[X[i]]:
                p[i] = np.random.uniform()
                h[i] = 0
            else:
                p[i] = np.random.beta(a = np.random.uniform(0.2,0.4), b = 4)
                h[i] = 1
        return p,h,X
   
    
def neuralfdr_generate_data_2D(job=0, n_samples=100000,data_vis=0):
    np.random.seed(42)
    if job == 0: # Gaussian mixtures 
        x1 = np.random.uniform(-1,1,size = n_samples)
        x2 = np.random.uniform(-1,1,size = n_samples)
        pi1 = ((mlab.bivariate_normal(x1, x2, 0.25, 0.25, -0.5, -0.2)+
               mlab.bivariate_normal(x1, x2, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)        
        p = np.zeros(n_samples)
        h = np.zeros(n_samples)
               
        for i in range(n_samples):
            rnd = np.random.uniform()
            if rnd > pi1[i]:
                p[i] = np.random.uniform()
                h[i] = 0
            else:
                p[i] = np.random.beta(a = 0.3, b = 4)
                h[i] = 1
        X = np.concatenate([[x1],[x2]]).T
        X = (X+1)/2
      
        if data_vis == 1:
            fig = plt.figure()
            ax1 = fig.add_subplot(121)
            x_grid = np.arange(-1, 1, 1/100.0)
            y_grid = np.arange(-1, 1, 1/100.0)
            X_grid, Y_grid = np.meshgrid(x_grid, y_grid)
            pi1_grid = ((mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, -0.5, -0.2)+
               mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)  
            ax1.pcolor(X_grid, Y_grid, pi1_grid)
            
            ax2 = fig.add_subplot(122)
            alt=ax2.scatter(x1[h==1][1:50], x2[h==1][1:50],color='r')
            nul=ax2.scatter(x1[h==0][1:50], x2[h==0][1:50],color='b')
            ax2.legend((alt,nul),('50 alternatives', '50 nulls'))
            
        return p, h, X
    if job == 1: # Linear trend
        
        x1 = np.random.uniform(-1,1,size = n_samples)
        x2 = np.random.uniform(-1,1,size = n_samples)
        pi1 = 0.1 * (x1 + 1) /2 +  0.3 *(1-x2) / 2
        
        p = np.zeros(n_samples)
        h = np.zeros(n_samples)
         
        for i in range(n_samples):
            rnd = np.random.uniform()
            if rnd > pi1[i]:
                p[i] = np.random.uniform()
                h[i] = 0
            else:
                p[i] = np.random.beta(a = 0.3, b = 4)
                h[i] = 1
        X = np.concatenate([[x1],[x2]]).T
        X = (X+1)/2
        
        if data_vis == 1:
            fig = plt.figure()
            ax1 = fig.add_subplot(121)
            x_grid = np.arange(-1, 1, 1/100.0)
            y_grid = np.arange(-1, 1, 1/100.0)
            X_grid, Y_grid = np.meshgrid(x_grid, y_grid)
            pi1_grid =  0.1 * (X_grid + 1) /2 +  0.3 *(1-Y_grid) / 2
            
            ax1.pcolor(X_grid, Y_grid, pi1_grid)
            
            ax2 = fig.add_subplot(122)
            alt=ax2.scatter(x1[h==1][1:50], x2[h==1][1:50],color='r')
            nul=ax2.scatter(x1[h==0][1:50], x2[h==0][1:50],color='b')
            ax2.legend((alt,nul),('50 alternatives', '50 nulls'))
            
        return p, h, X
        
        
        
    if job == 2: # Gaussian mixture + linear trend
        x1 = np.random.uniform(-1,1,size = n_samples)
        x2 = np.random.uniform(-1,1,size = n_samples)
        pi1 = ((mlab.bivariate_normal(x1, x2, 0.25, 0.25, -0.5, -0.2)+
               mlab.bivariate_normal(x1, x2, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)        
        pi1 = pi1 * 0.5 + 0.5*(0.5 * (x1 + 1) /2 +  0.3 *(1-x2) / 2)
        
        p = np.zeros(n_samples)
        h = np.zeros(n_samples)
               
        for i in range(n_samples):
            rnd = np.random.uniform()
            if rnd > pi1[i]:
                p[i] = np.random.uniform()
                h[i] = 0
            else:
                p[i] = np.random.beta(a = 0.3, b = 4)
                h[i] = 1
        X = np.concatenate([[x1],[x2]]).T
        X = (X+1)/2
        
        if data_vis == 1:
            fig = plt.figure()
            ax1 = fig.add_subplot(121)
            x_grid = np.arange(-1, 1, 1/100.0)
            y_grid = np.arange(-1, 1, 1/100.0)
            X_grid, Y_grid = np.meshgrid(x_grid, y_grid)
            pi1_grid = ((mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, -0.5, -0.2)+
               mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)  * 0.5 + (0.5 * (0.5 * (X_grid + 1) /2 +  0.3 *(1-Y_grid) / 2))
            ax1.pcolor(X_grid, Y_grid, pi1_grid)
            
            ax2 = fig.add_subplot(122)
            alt=ax2.scatter(x1[h==1][1:50], x2[h==1][1:50],color='r')
            nul=ax2.scatter(x1[h==0][1:50], x2[h==0][1:50],color='b')
            ax2.legend((alt,nul),('50 alternatives', '50 nulls'))
            
        return p, h, X

def load_2DGM(n_samples=100000,verbose=False):
    np.random.seed(42)
    x1 = np.random.uniform(-1,1,size = n_samples)
    x2 = np.random.uniform(-1,1,size = n_samples)
    pi1 = ((mlab.bivariate_normal(x1, x2, 0.25, 0.25, -0.5, -0.2)+
               mlab.bivariate_normal(x1, x2, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)        
    p = np.zeros(n_samples)
    h = np.zeros(n_samples)
               
    for i in range(n_samples):
        rnd = np.random.uniform()
        if rnd > pi1[i]:
            p[i] = np.random.uniform()
            h[i] = 0
        else:
            p[i] = np.random.beta(a = 0.3, b = 4)
            h[i] = 1
    X = np.concatenate([[x1],[x2]]).T
    X = (X+1)/2
      
    if verbose:
        fig = plt.figure()
        ax1 = fig.add_subplot(121)
        x_grid = np.arange(-1, 1, 1/100.0)
        y_grid = np.arange(-1, 1, 1/100.0)
        X_grid, Y_grid = np.meshgrid(x_grid, y_grid)
        pi1_grid = ((mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, -0.5, -0.2)+
           mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)  
        ax1.pcolor(X_grid, Y_grid, pi1_grid)
        
        ax2 = fig.add_subplot(122)
        alt=ax2.scatter(x1[h==1][1:50], x2[h==1][1:50],color='r')
        nul=ax2.scatter(x1[h==0][1:50], x2[h==0][1:50],color='b')
        ax2.legend((alt,nul),('50 alternatives', '50 nulls'))           
    return p,h,X

def load_2Dslope(n_samples=100000,verbose=False):
    x1 = np.random.uniform(-1,1,size = n_samples)
    x2 = np.random.uniform(-1,1,size = n_samples)
    pi1 = 0.1 * (x1 + 1) /2 +  0.3 *(1-x2) / 2
    
    p = np.zeros(n_samples)
    h = np.zeros(n_samples)
     
    for i in range(n_samples):
        rnd = np.random.uniform()
        if rnd > pi1[i]:
            p[i] = np.random.uniform()
            h[i] = 0
        else:
            p[i] = np.random.beta(a = 0.3, b = 4)
            h[i] = 1
    X = np.concatenate([[x1],[x2]]).T
    X = (X+1)/2
    
    if verbose:
        fig = plt.figure()
        ax1 = fig.add_subplot(121)
        x_grid = np.arange(-1, 1, 1/100.0)
        y_grid = np.arange(-1, 1, 1/100.0)
        X_grid, Y_grid = np.meshgrid(x_grid, y_grid)
        pi1_grid =  0.1 * (X_grid + 1) /2 +  0.3 *(1-Y_grid) / 2
        
        ax1.pcolor(X_grid, Y_grid, pi1_grid)
        
        ax2 = fig.add_subplot(122)
        alt=ax2.scatter(x1[h==1][1:50], x2[h==1][1:50],color='r')
        nul=ax2.scatter(x1[h==0][1:50], x2[h==0][1:50],color='b')
        ax2.legend((alt,nul),('50 alternatives', '50 nulls'))       
    return p,h,X

def load_2DGM_slope(n_samples=100000,verbose=False):
    x1 = np.random.uniform(-1,1,size = n_samples)
    x2 = np.random.uniform(-1,1,size = n_samples)
    pi1 = ((mlab.bivariate_normal(x1, x2, 0.25, 0.25, -0.5, -0.2)+
           mlab.bivariate_normal(x1, x2, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)        
    pi1 = pi1 * 0.5 + 0.5*(0.5 * (x1 + 1) /2 +  0.3 *(1-x2) / 2)
    
    p = np.zeros(n_samples)
    h = np.zeros(n_samples)
           
    for i in range(n_samples):
        rnd = np.random.uniform()
        if rnd > pi1[i]:
            p[i] = np.random.uniform()
            h[i] = 0
        else:
            p[i] = np.random.beta(a = 0.3, b = 4)
            h[i] = 1
    X = np.concatenate([[x1],[x2]]).T
    X = (X+1)/2
    
    if verbose:
        fig = plt.figure()
        ax1 = fig.add_subplot(121)
        x_grid = np.arange(-1, 1, 1/100.0)
        y_grid = np.arange(-1, 1, 1/100.0)
        X_grid, Y_grid = np.meshgrid(x_grid, y_grid)
        pi1_grid = ((mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, -0.5, -0.2)+
           mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)  * 0.5 + (0.5 * (0.5 * (X_grid + 1) /2 +  0.3 *(1-Y_grid) / 2))
        ax1.pcolor(X_grid, Y_grid, pi1_grid)
        
        ax2 = fig.add_subplot(122)
        alt=ax2.scatter(x1[h==1][1:50], x2[h==1][1:50],color='r')
        nul=ax2.scatter(x1[h==0][1:50], x2[h==0][1:50],color='b')
        ax2.legend((alt,nul),('50 alternatives', '50 nulls'))
    return p,h,X

def load_5DGM(n_sample=100000,verbose=False):
    p,h,x = load_2DGM(n_samples=n_sample,verbose=verbose) 
    x_noise = np.random.uniform(high=1,low=-1,size = (n_sample,3))
    x = np.concatenate([x,x_noise],1)
    return p,h,x

def load_100D(n_sample=100000,verbose=False):
    def generate_data_1D_cont(pi1,X):
        n_samples = len(X)
        p = np.zeros(n_samples)
        h = np.zeros(n_samples)
        
        for i in range(n_samples):
            rnd = np.random.uniform()
            if rnd > pi1[i]:
                p[i] = np.random.uniform()
                h[i] = 0
            else:
                p[i] = np.random.beta(a = np.random.uniform(0.2,0.4), b = 4)
                h[i] = 1
        return p, h, X
    
    X = np.random.uniform(high = 5, size = (n_sample,))
    pi1 = (5-X) / 10.0
    p, h, x = generate_data_1D_cont(pi1, X)
    x_noise = np.random.uniform(high = 5, size = (n_sample,99))
    x = np.concatenate([np.expand_dims(x,1), x_noise], 1)
    return p,h,x

def load_airway(verbose=False):
    file_name='/data3/martin/nfdr2_simulation_data/RNA_seq/airway'
    X = np.loadtxt(file_name,skiprows=0,delimiter=',')
    x=X[:,2].reshape([-1,1])
    p=X[:,0]
    return p, None, x

def load_bottomly(verbose=False):
    file_name='/data3/martin/nfdr2_simulation_data/RNA_seq/bottomly'
    X = np.loadtxt(file_name,skiprows=0,delimiter=',')
    x=X[:,2].reshape([-1,1])
    p=X[:,0]
    return p, None, x

def load_pasilla(verbose=False):
    file_name='/data3/martin/nfdr2_simulation_data/RNA_seq/pasilla'
    X = np.loadtxt(file_name,skiprows=0,delimiter=',')
    x=X[:,2].reshape([-1,1])
    p=X[:,0]
    return p, None, x

def load_proteomics(verbose=False):
    file_name='/data/martin/NeuralFDR/NeuralFDR_data/proteomics.csv'
    X = np.loadtxt(file_name,skiprows=0,delimiter=',')
    x=X[:,0]
    p=X[:,1]    
    if verbose:
        print('## proteomics.csv ##')
        print('# hypothesis: %s'%str(x.shape[0]))
        for i in range(5):
            print('p=%s, x=%s'%(str(p[i]),str(x[i])))
        print('\n')
    return p,x

def load_GTEx_1d(verbose=False):
    file_name='/data/martin/NeuralFDR/NeuralFDR_data/data_gtex.csv'
    X = np.loadtxt(file_name,skiprows=1,delimiter=',')
    x=X[:,0]
    p=X[:,1]    
    if verbose:
        print('## airway data ##')
        print('# hypothesis: %s'%str(x.shape[0]))
        for i in range(5):
            print('p=%s, x=%s'%(str(p[i]),str(x[i])))
        print('\n')
    return p,x

"""
    Load the GTEx full data. 
    Data are only kept for those with p-values >0.995 or <0.005.
    The full data size is 10623893
""" 
def load_GTEx_full(verbose=False):
    file_name='/data/martin/NeuralFDR/NeuralFDR_data/gtex_new_filtered.csv'
    X = np.loadtxt(file_name,skiprows=1,delimiter=',')
    x,p,n_full = X[:,0:4],X[:,4],10623893
    #x[:,0],x[:,1] = np.log(x[:,0]+1), np.log(x[:,1]+1)
    if verbose:
        print('## Load GTEx full data ##')
        print('# all hypothesis: %d'%n_full)
        print('# filtered hypothesis: %d'%x.shape[0])
        for i in range(5):
            print('# p=%s, x=%s'%(str(p[i]),str(x[i])))
        print('\n')
        
    cate_name = {'Art': 0, 'Ctcf': 1, 'CtcfO': 2, 'DnaseD': 3, 'DnaseU': 4, 'Elon': 5, 'ElonW': 6, 'Enh': 7, 'EnhF': 8, 'EnhW': 9, 'EnhWF': 10, 'FaireW': 11, "Gen3'": 12, "Gen5'": 13, 'H4K20': 14, 'Low': 15, 'Pol2': 16, 'PromF': 17, 'PromP': 18, 'Quies': 19, 'Repr': 20, 'ReprD': 21, 'ReprW': 22, 'Tss': 23, 'TssF': 24}
    cate_name = {v: k for k, v in cate_name.items()}
    
    cate_name_dic = {}
    cate_name_dic[3] = cate_name
    #cate_name = [None,None,None,cate_name]
    return p,x,n_full,cate_name_dic

def load_GTEx_small():
    n_full = 172353475
    fpath = '/data3/martin/gtex_data/GTEx_Analysis_v7_eQTL_all_associations'
    fname = 'GTEx_small.pickle'
    fname = fpath + '/' + fname
    with open(fname, 'rb') as handle:  
        p = pickle.load(handle)
        x = pickle.load(handle)
        n_full = pickle.load(handle)
    return p, x, n_full, {}

def load_GTEx_Adipose_Subcutaneous():
    """ Load data for Adipose_Subcutaneous
    """
    n_full = 172353475
    fpath = '/data3/martin/gtex_data/GTEx_Analysis_v7_eQTL_all_associations'
    fname = 'Adipose_Subcutaneous.allpairs.txt.processed.filtered'
    fname = fpath + '/' + fname
    data = np.loadtxt(fname, dtype=str, delimiter=', ')
    hypothesis_name = data[:, 0]
    p = data[:, -1].astype(dtype = float)
    x = data[:, 1:5].astype(dtype = float)
    ind_nan = np.isnan(x[:, 1])
    x[ind_nan, 1] = np.mean(x[~ind_nan, 1])
    x = x[p<1, :]
    p = p[p<1]
    return p, x, n_full, {}

def load_GTEx_Colon_Sigmoid():
    """ Load data for Colon_Sigmoid
    """
    n_full = 170481049
    fpath = '/data3/martin/gtex_data/GTEx_Analysis_v7_eQTL_all_associations'
    fname = 'Colon_Sigmoid.allpairs.txt.processed.filtered'
    fname = fpath + '/' + fname
    data = np.loadtxt(fname, dtype=str, delimiter=', ')
    hypothesis_name = data[:, 0]
    p = data[:, -1].astype(dtype = float)
    x = data[:, 1:5].astype(dtype = float)
    ind_nan = np.isnan(x[:, 1])
    x[ind_nan, 1] = np.mean(x[~ind_nan, 1])    
    x = x[p<1, :]
    p = p[p<1]
    return p, x, n_full, {}

def load_GTEx_Artery_Aorta():
    """ Load data for Artery_Aorta
    """
    n_full = 166456366
    fpath = '/data3/martin/gtex_data/GTEx_Analysis_v7_eQTL_all_associations'
    fname = 'Artery_Aorta.allpairs.txt.processed.filtered'
    fname = fpath + '/' + fname
    data = np.loadtxt(fname, dtype=str, delimiter=', ')
    hypothesis_name = data[:, 0]
    p = data[:, -1].astype(dtype = float)
    x = data[:, 1:5].astype(dtype = float)
    ind_nan = np.isnan(x[:, 1])
    x[ind_nan, 1] = np.mean(x[~ind_nan, 1])
    x = x[p<1, :]
    p = p[p<1]
    return p, x, n_full, {}

def load_GTEx(data_name='GTEx_small', if_impute=True):
    """ Load data for the GTEx data
    
    Data information:
        fixit: add the data information here.
    """
    print('load %s'%data_name)
    n_trunc = 300000
    # Hard-coded information of the GTEx dataset.
    cate_name = {3: {1: 'TssA', 2: 'TssAFlnk', 3: 'TxFlnk', 4: 'Tx',
                     5: 'TxWk', 6: 'EnhG', 7: 'Enh', 8: 'ZNF/Rpts',
                     9: 'Het', 10: 'TssBiv', 11: 'BivFlnk', 12: 'EnhBiv',
                     13: 'ReprPC', 14: 'ReprPCWk', 15: 'Quies'}}   
    dic_n_full = {'GTEx_test': 1,\
                  'test-aug': 1,\
                  'GTEx_small': 172353475,\
                  'Adipose_Subcutaneous': 172353475,\
                  'Adipose_Subcutaneous-aug': 172353475,\
                  'Adipose_Subcutaneous-a_ur': 172353475,\
                  'Adipose_Visceral_Omentum': 172595476,\
                  'Adipose_Visceral_Omentum-aug': 172595476,\
                  'Adipose_Visceral_Omentum-a_ur': 172595476,\
                  'Artery_Aorta': 166456366,\
                  'Breast_Mammary_Tissue': 179856829,\
                  'Cells_EBV-transformed_lymphocytes': 159717963,\
                  'Colon_Sigmoid': 170481049,\
                  'Colon_Sigmoid-aug': 170481049,\
                  'Colon_Sigmoid-a_ur': 170481049,\
                  'Colon_Transverse': 176504796,\
                  'Colon_Transverse-aug': 176504796,\
                  'Colon_Transverse-a_ur': 176504796,\
                  'Esophagus_Gastroesophageal_Junction': 167240321,\
                  'Esophagus_Mucosa': 167425246,\
                  'Esophagus_Muscularis': 165726398,\
                  'Heart_Atrial_Appendage': 161328107,\
                  'Heart_Left_Ventricle': 149705855,\
                  'Lung': 182309377,\
                  'Muscle_Skeletal': 147337341,\
                  'Pancreas': 158693140,\
                  'Stomach': 168679055,\
                  'Whole_Blood': 144733342,\
                  'Adipose_Subcutaneous-chr21': n_trunc,\
                  'Adipose_Visceral_Omentum-chr21': n_trunc}
    data_name_list = dic_n_full.keys()
    fpath = '/data3/martin/gtex_data/GTEx_Analysis_v7_eQTL_all_associations'
    # Load the information
    n_full = dic_n_full[data_name]
    if data_name == 'GTEx_small':
        fname = 'GTEx_small.pickle'
        fname = fpath + '/' + fname
        with open(fname, 'rb') as handle:  
            p = pickle.load(handle)
            x = pickle.load(handle)
            cis_name = pickle.load(handle)
    else:
        suffix = ''
        if 'chr21' in data_name:
            data_name, suffix = data_name.split('-')
            fname = data_name + '.allpairs.txt.processed.chr21.txt'
        elif 'aug' in data_name:
            data_name, suffix = data_name.split('-')
            fname = data_name + '.allpairs.txt.processed.filtered.augmented.txt'
        elif 'a_ur' in data_name:
            data_name, suffix = data_name.split('-')
            fname = data_name + '.allpairs.txt.processed.filtered.augmented_not_related.txt'
        else:            
            fname = data_name + '.allpairs.txt.processed.filtered'
        fname = fpath + '/' + fname
        data = np.loadtxt(fname, dtype=str, delimiter=', ')
        hypothesis_name = data[:, 0]        
        if (suffix == 'aug') or (suffix == 'a_ur'):
            x = data[:, [1,2,3,4,7]].astype(dtype = float)
            x[:, 4] = -np.log10(x[:, 4])
            p = data[:, -2].astype(dtype = float)
            cis_name = data[:, 0]
        elif suffix == 'chr21':
            x = data[:n_trunc, [1,2,3,4]].astype(dtype = float)
            p = data[:n_trunc, -1].astype(dtype = float)
            cis_name = data[:n_trunc, 0]
        else:
            x = data[:, [1,2,3,4]].astype(dtype = float)
            p = data[:, -1].astype(dtype = float)
            cis_name = data[:, 0]
        # nan values.
        if if_impute:
            for i in range(x.shape[1]):
                ind_nan = np.isnan(x[:, i])
                x[ind_nan, i] = np.mean(x[~ind_nan, i])
        else:
            # remove the nan values
            ind_nan = np.zeros([x.shape[0]], dtype=bool)
            for i in range(x.shape[1]):
                ind_nan[np.isnan(x[:, i])] = True
            x = x[~ind_nan, :]
            p = p[~ind_nan]
            cis_name = cis_name[~ind_nan]
        # Expression level.
        x[:, 0] = np.log10(x[:, 0]+0.5)
        x = x[p<1, :]
        cis_name = cis_name[p<1]
        p = p[p<1]      
    return p, x, n_full, cate_name, cis_name

""" 
    load ukbb breast cancer 
""" 
def load_ukbb_breast_cancer(verbose=False, use_other=False):
    file_name='/data/ukbb_process/breast_cancer_filtered.csv'
    file_name='/data/martin/breast_cancer_filtered.csv'
    X = np.loadtxt(file_name,skiprows=1,delimiter=',')
    if not use_other:
        x,p,n_full = X[:,0:2],X[:,-2],847800
    else:
        x,p,n_full = X[:,0:6],X[:,-2],847800
    
    #x[:,0],x[:,1] = np.log(x[:,0]+1), np.log(x[:,1]+1)
    if verbose:
        print('## Load ukbb breast cancer data ##')
        print('# all hypothesis: %d'%n_full)
        print('# filtered hypothesis: %d'%x.shape[0])
        for i in range(5):
            print('# p=%s, x=%s'%(str(p[i]),str(x[i])))
        print('')
        
    cate_name = {'Art': 0, 'Ctcf': 1, 'CtcfO': 2, 'DnaseD': 3, 'DnaseU': 4, 'Elon': 5, 'ElonW': 6, 'Enh': 7, 'EnhF': 8, 'EnhW': 9, 'EnhWF': 10, 'FaireW': 11, "Gen3'": 12, "Gen5'": 13, 'H4K20': 14, 'Low': 15, 'Pol2': 16, 'PromF': 17, 'PromP': 18, 'Quies': 19, 'Repr': 20, 'ReprD': 21, 'ReprW': 22, 'Tss': 23, 'TssF': 24}
    cate_name = {v: k for k, v in cate_name.items()}
    
    cate_name_dic = {}
    cate_name_dic[3] = cate_name
    #if not use_other:
    #    cate_name = [cate_name,None]
    #else:
    #    cate_name = [cate_name,None, None, None, None, None]
        
    return p,x,n_full,cate_name_dic

def load_common_dataset(filename,n,verbose=True):
    X = np.loadtxt(filename, skiprows=1, delimiter=',')
    x,p,n_full = X[:, 0:-2], X[:, -2], n
    #cat_name = [None] * (x.shape[1])
    cat_name = {}
    
    if verbose:
        print('## Load ukbb %s ##'%filename)
        print('# all hypothesis: %d'%n_full)
        print('# filtered hypothesis: %d'%x.shape[0])
        for i in range(5):
            print('# p=%s, x=%s'%(str(p[i]),str(x[i])))
        print('')
        
    return p, x, n_full, cat_name


import matplotlib
matplotlib.use('Agg') 
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import mlab

from scipy.stats import beta

import torch
import torch.nn as nn
from torch.autograd import Variable
import torch.optim as optim

from numpy import array
from scipy.cluster.vq import kmeans2
from sklearn.cluster import KMeans


def generate_data_1D(job=0, n_samples=10000,data_vis=0, num_case=4):
    if job == 0: # discrete case
        pi1=np.random.uniform(0,0.3,size=num_case)
        X=np.random.randint(0, num_case, n_samples)
        
        p = np.zeros(n_samples)
        h = np.zeros(n_samples)
        
        for i in range(n_samples):
            rnd = np.random.uniform()
            if rnd > pi1[X[i]]:
                p[i] = np.random.uniform()
                h[i] = 0
            else:
                p[i] = np.random.beta(a = np.random.uniform(0.2,0.4), b = 4)
                h[i] = 1
        return p, h, X

    
    
def generate_data_1D_cont(pi1, X, job=0):
    if job == 0: # discrete case
        n_samples = len(X)
        p = np.zeros(n_samples)
        h = np.zeros(n_samples)
        
        for i in range(n_samples):
            rnd = np.random.uniform()
            if rnd > pi1[i]:
                p[i] = np.random.uniform()
                h[i] = 0
            else:
                p[i] = np.random.beta(a = np.random.uniform(0.2,0.4), b = 4)
                h[i] = 1
        return p, h, X
    
    
#def p_value_beta_fit(p, lamb=0.8, bin_num=50, vis=0):
#    pi_0=np.divide(np.sum(p>lamb), p.shape[0] * (1-lamb))
#    temp_p=np.zeros([0])
#    step_size=np.divide(1,np.float(bin_num))
#    fil_num=np.int(np.divide(pi_0*p.shape[0],bin_num))+1
#    for i in range(bin_num):
#        p1=p[p>step_size*(i-1)]
#        p1=p1[p1 <= step_size*i]
#        choice_num= np.max(p1.shape[0] - fil_num,0)
#        if choice_num > 1:
#            choice=np.random.choice(p1.shape[0], choice_num)
#            temp_p=np.concatenate([temp_p,p1[choice]]).T
#    if vis==1:
#        plt.figure()
#        plt.hist(temp_p, bins=100, normed=True)       
#    a, b, loc, scale = beta.fit(temp_p,floc=0,fscale=1)
#    return pi_0, a, b

#def beta_mixture_pdf(x,pi_0,a,b):
#    return beta.pdf(x,a,b)*(1-pi_0)+pi_0
#
#def Storey_BH(x, alpha = 0.05, lamb=0.4, n = None):
#    pi0_hat=np.divide(np.sum(x>lamb),x.shape[0] *(1-lamb))
#    alpha /= pi0_hat
#    x_s = sorted(x)
#    if n is None:
#        n = len(x_s)
#    ic = 0
#    for i in range(len(x_s)):
#        if x_s[i] < i*alpha/float(n):
#            ic = i
#    return ic, x_s[ic], pi0_hat

#def Opt_t_cal_discrete(p, X, num_case=2,step_size=0.0001,ub=0.05,n_samples=10000,alpha=0.05):
#    # Fit the beta mixture parameters
#    fit_param=np.zeros([num_case, 3])
#    for i in range(num_case):
#        fit_param[i,:]=p_value_beta_fit(p[X==i])
#
#    # Calculating the ratios 
#    t_opt=np.zeros([num_case])
#    max_idx=np.argmin(fit_param[:,0])
#    x_grid = np.arange(0, ub, step_size)
#    t_ratio=np.zeros([num_case,x_grid.shape[0]])
#    for i in range(num_case):
#        t_ratio[i,:] = np.divide(beta_mixture_pdf(
#            x_grid,fit_param[i,0],fit_param[i,1],fit_param[i,2]), fit_param[i,0])
#
#    # Increase the threshold
#    for i in range(len(x_grid)):
#        t=np.zeros([num_case])
#        # undate the search optimal threshold
#        t[max_idx]=x_grid[i]
#        c=t_ratio[max_idx,i]
#        for j in range(num_case):
#            if j != max_idx: 
#                for k in range(len(x_grid)):
#                    if k == len(x_grid)-1:
#                        t[j]=x_grid[k]
#                        break
#                    if t_ratio[j,k+1]<c:
#                        t[j]=x_grid[k]
#                        break
#        # calculate the FDR
#        num_dis=0 
#        num_fd =0 
#        for i in range(num_case):
#            num_dis+=np.sum(p[X==i] < t[i])
#            num_fd+=np.sum(X==i)*t[i]*fit_param[i,0]
#
#        if np.divide(num_fd,np.float(np.amax([num_dis,1])))<alpha:
#            t_opt=t
#        else:
#            break
#    return t_opt

def generate_data_2D(job=0, n_samples=10000,data_vis=0):
    if job == 0: # Gaussian mixtures 
        x1 = np.random.uniform(-1,1,size = n_samples)
        x2 = np.random.uniform(-1,1,size = n_samples)
        pi1 = ((mlab.bivariate_normal(x1, x2, 0.25, 0.25, -0.5, -0.2)+
               mlab.bivariate_normal(x1, x2, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)        
        p = np.zeros(n_samples)
        h = np.zeros(n_samples)
               
        for i in range(n_samples):
            rnd = np.random.uniform()
            if rnd > pi1[i]:
                p[i] = np.random.uniform()
                h[i] = 0
            else:
                p[i] = np.random.beta(a = 0.3, b = 4)
                h[i] = 1
        X = np.concatenate([[x1],[x2]]).T
        
        if data_vis == 1:
            fig = plt.figure()
            ax1 = fig.add_subplot(121)
            x_grid = np.arange(-1, 1, 1/100.0)
            y_grid = np.arange(-1, 1, 1/100.0)
            X_grid, Y_grid = np.meshgrid(x_grid, y_grid)
            pi1_grid = ((mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, -0.5, -0.2)+
               mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)  
            ax1.pcolor(X_grid, Y_grid, pi1_grid)
            
            ax2 = fig.add_subplot(122)
            alt=ax2.scatter(x1[h==1][1:50], x2[h==1][1:50],color='r')
            nul=ax2.scatter(x1[h==0][1:50], x2[h==0][1:50],color='b')
            ax2.legend((alt,nul),('50 alternatives', '50 nulls'))
            
        return p, h, X
    if job == 1: # Linear trend
        
        x1 = np.random.uniform(-1,1,size = n_samples)
        x2 = np.random.uniform(-1,1,size = n_samples)
        pi1 = 0.1 * (x1 + 1) /2 +  0.3 *(1-x2) / 2
        
        p = np.zeros(n_samples)
        h = np.zeros(n_samples)
         
        for i in range(n_samples):
            rnd = np.random.uniform()
            if rnd > pi1[i]:
                p[i] = np.random.uniform()
                h[i] = 0
            else:
                p[i] = np.random.beta(a = 0.3, b = 4)
                h[i] = 1
        X = np.concatenate([[x1],[x2]]).T
        
        if data_vis == 1:
            fig = plt.figure()
            ax1 = fig.add_subplot(121)
            x_grid = np.arange(-1, 1, 1/100.0)
            y_grid = np.arange(-1, 1, 1/100.0)
            X_grid, Y_grid = np.meshgrid(x_grid, y_grid)
            pi1_grid =  0.1 * (X_grid + 1) /2 +  0.3 *(1-Y_grid) / 2
            
            ax1.pcolor(X_grid, Y_grid, pi1_grid)
            
            ax2 = fig.add_subplot(122)
            alt=ax2.scatter(x1[h==1][1:50], x2[h==1][1:50],color='r')
            nul=ax2.scatter(x1[h==0][1:50], x2[h==0][1:50],color='b')
            ax2.legend((alt,nul),('50 alternatives', '50 nulls'))
            
        return p, h, X
        
        
        
    if job == 2: # Gaussian mixture + linear trend
        x1 = np.random.uniform(-1,1,size = n_samples)
        x2 = np.random.uniform(-1,1,size = n_samples)
        pi1 = ((mlab.bivariate_normal(x1, x2, 0.25, 0.25, -0.5, -0.2)+
               mlab.bivariate_normal(x1, x2, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)        
        pi1 = pi1 * 0.5 + 0.5*(0.5 * (x1 + 1) /2 +  0.3 *(1-x2) / 2)
        
        p = np.zeros(n_samples)
        h = np.zeros(n_samples)
               
        for i in range(n_samples):
            rnd = np.random.uniform()
            if rnd > pi1[i]:
                p[i] = np.random.uniform()
                h[i] = 0
            else:
                p[i] = np.random.beta(a = 0.3, b = 4)
                h[i] = 1
        X = np.concatenate([[x1],[x2]]).T
        
        if data_vis == 1:
            fig = plt.figure()
            ax1 = fig.add_subplot(121)
            x_grid = np.arange(-1, 1, 1/100.0)
            y_grid = np.arange(-1, 1, 1/100.0)
            X_grid, Y_grid = np.meshgrid(x_grid, y_grid)
            pi1_grid = ((mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, -0.5, -0.2)+
               mlab.bivariate_normal(X_grid, Y_grid, 0.25, 0.25, 0.7, 0.5))/2).clip(max=1)  * 0.5 + (0.5 * (0.5 * (X_grid + 1) /2 +  0.3 *(1-Y_grid) / 2))
            ax1.pcolor(X_grid, Y_grid, pi1_grid)
            
            ax2 = fig.add_subplot(122)
            alt=ax2.scatter(x1[h==1][1:50], x2[h==1][1:50],color='r')
            nul=ax2.scatter(x1[h==0][1:50], x2[h==0][1:50],color='b')
            ax2.legend((alt,nul),('50 alternatives', '50 nulls'))
            
        return p, h, X
    
#def BH(x, alpha = 0.05, n = None):
#    x_s = sorted(x)
#    if n is None:
#        n = len(x_s)
#    ic = 0
#    for i in range(len(x_s)):
#        if x_s[i] < i*alpha/float(n):
#            ic = i
#    return ic, x_s[ic]



#def p_value_beta_fit(p, lamb=0.8, bin_num=50, vis=0):
#    pi_0=np.divide(np.sum(p>lamb), p.shape[0] * (1-lamb))
#    temp_p=np.zeros([0])
#    step_size=np.divide(1,np.float(bin_num))
#    fil_num=np.int(np.divide(pi_0*p.shape[0],bin_num))+1
#    for i in range(bin_num):
#        p1=p[p>step_size*(i-1)]
#        p1=p1[p1 <= step_size*i]
#        choice_num= np.max(p1.shape[0] - fil_num,0)
#        if choice_num > 1:
#            choice=np.random.choice(p1.shape[0], choice_num)
#            temp_p=np.concatenate([temp_p,p1[choice]]).T
#    if vis==1:
#        plt.figure()
#        plt.hist(temp_p, bins=100, normed=True)       
#    a, b, loc, scale = beta.fit(temp_p,floc=0,fscale=1)
#    return pi_0, a, b
#def beta_mixture_pdf(x,pi_0,a,b):
#    return beta.pdf(x,a,b)*(1-pi_0)+pi_0
#


#def Opt_t_cal_discrete(p, X, num_case=2,step_size=0.0001,ub=0.05,n_samples=10000,alpha=0.05):
#    # Fit the beta mixture parameters
#    fit_param=np.zeros([num_case, 3])
#    for i in range(num_case):
#        fit_param[i,:]=p_value_beta_fit(p[X==i])
#
#    # Calculating the ratios 
#    t_opt=np.zeros([num_case])
#    max_idx=np.argmin(fit_param[:,0])
#    x_grid = np.arange(0, ub, step_size)
#    t_ratio=np.zeros([num_case,x_grid.shape[0]])
#    for i in range(num_case):
#        t_ratio[i,:] = np.divide(beta_mixture_pdf(
#            x_grid,fit_param[i,0],fit_param[i,1],fit_param[i,2]), fit_param[i,0])
#
#    # Increase the threshold
#    for i in range(len(x_grid)):
#        t=np.zeros([num_case])
#        # undate the search optimal threshold
#        t[max_idx]=x_grid[i]
#        c=t_ratio[max_idx,i]
#        for j in range(num_case):
#            if j != max_idx: 
#                for k in range(len(x_grid)):
#                    if k == len(x_grid)-1:
#                        t[j]=x_grid[k]
#                        break
#                    if t_ratio[j,k+1]<c:
#                        t[j]=x_grid[k]
#                        break
#        # calculate the FDR
#        num_dis=0 
#        num_fd =0 
#        for i in range(num_case):
#            num_dis+=np.sum(p[X==i] < t[i])
#            num_fd+=np.sum(X==i)*t[i]*fit_param[i,0]
#
#        if np.divide(num_fd,np.float(np.amax([num_dis,1])))<alpha:
#            t_opt=t
#        else:
#            break
#    return t_opt

def result_summary(h,pred):
    print("Num of alternatives:",np.sum(h))
    print("Num of discovery:",np.sum(pred))
    print("Num of true discovery:",np.sum(pred * h))
    print("Actual FDR:", 1-np.sum(pred * h) / np.sum(pred))
    
#def softmax_prob_cal(X,Centorid, intensity=1):
#    dist=np.zeros([n_samples,num_clusters])
#    dist+=np.sum(X*X,axis=1, keepdims=True)
#    dist+=np.sum(centroid.T*centroid.T,axis=0, keepdims=True)
#    dist -= 2*X.dot(centroid.T)
#    dist=np.exp(dist*intensity)
#    dist /= np.sum(dist,axis=1, keepdims=True)
#    return dist


#def get_network(num_layers = 10, node_size = 10, dim = 1, scale = 1, cuda = False):
#    
#    
#    class Model(nn.Module):
#        def __init__(self, num_layers, node_size, dim):
#            super(Model, self).__init__()
#            l = []
#            l.append(nn.Linear(dim,node_size))
#            l.append(nn.LeakyReLU(0.1))
#            for i in range(num_layers - 2):
#                l.append(nn.Linear(node_size,node_size))
#                l.append(nn.LeakyReLU(0.1))
#
#            l.append(nn.Linear(node_size,1))
#            l.append(nn.Sigmoid())
#
#            self.layers = nn.Sequential(*l)
#
#
#
#        def forward(self, x):
#            x = self.layers(x)
#            x = 0.5 * scale * x 
#            return x
#
#   
#    
#    
#    network = Model(num_layers, node_size, dim)
#    if cuda:
#        return network.cuda()
#    else:
#        return network


#def train_network_to_target_p(network, optimizer, x, target_p, num_it = 1000, dim = 1, cuda = False):
#    target = Variable(torch.from_numpy(target_p.astype(np.float32)))
#    l1loss = nn.L1Loss()
#    batch_size = len(x)
#    n_samples = len(x)
#    loss_hist = []
#    choice = range(n_samples)
#    x_input = Variable(torch.from_numpy(x[choice].astype(np.float32).reshape(batch_size,dim)))
#
#    
#    if cuda:
#        x_input = x_input.cuda()
#        target = target.cuda()
#
#    for iteration in range(num_it):
#        if iteration % 100 == 0:
#            print iteration
#        
#        optimizer.zero_grad()
#        output = network.forward(x_input) 
#
#        loss = l1loss(output, target)
#        loss.backward()
#
#        optimizer.step()
#        loss_hist.append(loss.data[0])
#    
#    return loss_hist
#
#def train_network(network, optimizer, x, p, num_it = 3000, alpha = 0.05, dim = 1, lambda_ = 20, lambda2_ = 1e3, cuda = False, fdr_scale = 1, mirror = 1):
#    
#    batch_size = len(x)
#    n_samples = len(x)
#    print(batch_size, n_samples)
#    loss_hist = []
#    soft_compare = nn.Sigmoid()
#
#    relu = nn.ReLU()
#    choice = range(n_samples)
#    x_input = Variable(torch.from_numpy(x[choice].astype(np.float32).reshape(batch_size,dim)))
#    p_input = Variable(torch.from_numpy(p[choice].astype(np.float32).reshape(batch_size,1)))
#
#    if cuda:
#        x_input = x_input.cuda()
#        p_input = p_input.cuda()
#        
#    for iteration in range(num_it):
#        if iteration % 100 == 0:
#            print iteration
#
#
#        optimizer.zero_grad()
#        output = network.forward(x_input) 
#        s = torch.sum(soft_compare((output - p_input) * lambda2_)) / batch_size #disco rate
#        s2 = torch.sum(soft_compare((p_input - (mirror - output * fdr_scale)) * lambda2_)) / batch_size /float(fdr_scale)#false discoverate rate(over all samples)
#
#        gain = s  - lambda_ * relu((s2 - s * alpha)) 
#
#        loss = -gain
#        loss.backward()
#
#        optimizer.step()
#        loss_hist.append(loss.data[0])
#    
#    return loss_hist, s, s2
#
#
##def train_network_adapt(network, optimizer, x, p, num_it = 3000, alpha = 0.05, dim = 1, lambda_ = 2, lambda2_ = 1e3, cuda = False):
##    
##    batch_size = len(x)
##    n_samples = len(x)
##    loss_hist = []
##    soft_compare = nn.Sigmoid()
##
##    relu = nn.ReLU()
##    choice = range(n_samples)
##    x_input = Variable(torch.from_numpy(x[choice].astype(np.float32).reshape(batch_size,dim)))
##    p_input = Variable(torch.from_numpy(p[choice].astype(np.float32).reshape(batch_size,1)))
##
##    if cuda:
##        x_input = x_input.cuda()
##        p_input = p_input.cuda()
##        
##    for iteration in range(num_it):
##        if iteration % 100 == 0:
##            print iteration
##
##
##        optimizer.zero_grad()
##        output = network.forward(x_input) 
##        s = torch.sum(soft_compare((output - p_input) * lambda2_)) / batch_size #disco rate
##        s2 = torch.sum(soft_compare((p_input - (1-output)) * lambda2_)) / batch_size #false discoverate rate(over all samples)
##
##        gain = s  - lambda_ * (1+iteration/1000.0) * relu((s2 - s * alpha)) 
##
##        loss = -gain
##        loss.backward()
##
##        optimizer.step()
##        loss_hist.append(loss.data[0])
##    
##    return loss_hist, s, s2
#
#
#
#def opt_threshold(x, p, k, intensity = 1):
#    n_samples = x.shape[0]
#    
#    if len(x.shape) == 1:
#        x = np.expand_dims(x,1)
#    km = KMeans(n_clusters = k)
#    cluster = km.fit_predict(x)
#    opt = Opt_t_cal_discrete(p, cluster, num_case = k, step_size=0.00001)
#    #p_target = opt[cluster]
#    
#    x2 = x.repeat(k, axis = 1)
#    center = km.cluster_centers_.repeat(n_samples, axis = 1).T
#
#    e = np.exp (- (x2 - center) ** 2 / intensity)
#    s = np.expand_dims(np.sum(e, axis = 1),1)
#    prob = e/s
#    opt = Opt_t_cal_discrete(p, cluster, num_case = 10, step_size=0.00001)
#    p_target = prob.dot(opt)
#    
#    return p_target
#
#
#def opt_threshold_multi(x, p, k, intensity = 1, alpha = 0.05):
#    n_samples = x.shape[0]
#
#    km = KMeans(n_clusters = k)
#    cluster = km.fit_predict(x)
#    opt = Opt_t_cal_discrete(p, cluster, num_case = k, step_size=0.00001, alpha = alpha)
#    center = np.expand_dims(km.cluster_centers_, axis = -1).repeat(n_samples, axis = -1).T
#    x2 = np.expand_dims(x, axis = -1).repeat(k, axis = -1)
#
#    dist = x2 - center
#    dist = np.sum((x2 - center) ** 2, axis = 1)
#
#
#    e = np.exp (- dist / intensity)
#    s = np.expand_dims(np.sum(e, axis = 1),1)
#    prob = e/s
#    p_target = prob.dot(opt)
#    
#    return p_target
#
#
#
#def train_network_val(network, optimizer, x, p, num_it = 3000, alpha = 0.05, dim = 1, lambda_ = 20, lambda2_ = 1e3, lambda3_ = 1e2, cuda = False):
#    
#    batch_size = len(x)
#    n_samples = len(x)
#    train_loss_hist = []
#    val_loss_hist = []
#    soft_compare = nn.Sigmoid()
#
#    relu = nn.ReLU()
#    train_idx = range(n_samples/2)
#    val_idx = range(n_samples/2, n_samples)
#    num_train = len(train_idx)
#    num_val = len(val_idx)
#    
#    x_input_train = Variable(torch.from_numpy(x[train_idx].astype(np.float32).reshape(num_train,dim)))
#    p_input_train = Variable(torch.from_numpy(p[train_idx].astype(np.float32).reshape(num_train,1)))
#    x_input_val = Variable(torch.from_numpy(x[val_idx].astype(np.float32).reshape(num_val,dim)))
#    p_input_val = Variable(torch.from_numpy(p[val_idx].astype(np.float32).reshape(num_val,1)))
#    
#
#    if cuda:
#        x_input_train = x_input_train.cuda()
#        p_input_train = p_input_train.cuda()
#        x_input_val = x_input_val.cuda()
#        p_input_val = p_input_val.cuda()
#        
#    for iteration in range(num_it):
#        if iteration % 100 == 0:
#            print iteration
#
#
#        optimizer.zero_grad()
#        output = network.forward(x_input_train) 
#        s = torch.sum(soft_compare((output - p_input_train) * lambda2_)) / num_train #disco rate
#        s2 = torch.sum(soft_compare((p_input_train - (1-output)) * lambda3_)) / num_train #false discoverate rate(over all samples)
#
#        gain = s  - lambda_ * relu((s2 - s * alpha)) 
#
#        loss = -gain
#        loss.backward()
#
#        optimizer.step()
#        train_loss_hist.append(loss.data[0])
#        
#        output = network.forward(x_input_val) 
#        s = torch.sum(soft_compare((output - p_input_val) * lambda2_)) / num_val #disco rate
#        s2 = torch.sum(soft_compare((p_input_val - (1-output)) * lambda3_)) / num_val #false discoverate rate(over all samples)
#
#        gain = s  - lambda_ * relu((s2 - s * alpha)) 
#
#        loss = -gain
#        val_loss_hist.append(loss.data[0])
#        
#    
#    return train_loss_hist, val_loss_hist, s, s2
#
#
#def get_scale(network, x, p, dim = 1, cuda = False, lambda_ = 20, lambda2_ = 1e3, alpha = 0.05, fit = False, scale = 1, fdr_scale = 1, mirror  = 1):
#    batch_size = len(x)
#    n_samples = len(x)
#    loss_hist = []
#    soft_compare = nn.Sigmoid()
#    
#    relu = nn.ReLU()
#    choice = range(n_samples)
#    x_input = Variable(torch.from_numpy(x[choice].astype(np.float32).reshape(batch_size,dim)))
#    p_input = Variable(torch.from_numpy(p[choice].astype(np.float32).reshape(batch_size,1)))
#
#    if cuda:
#        x_input = x_input.cuda()
#        p_input = p_input.cuda()
#    
#    hi = 10.0
#    low = 0.1
#    current = scale
#    output = network.forward(x_input) * current
#
#    s = torch.sum(soft_compare((output - p_input) * lambda2_)) / batch_size #disco rate
#    s2 = torch.sum(soft_compare((p_input - (mirror-output)) * lambda2_)) / batch_size #false discoverate rate(over all samples)
#
#
#            
#    if fit:
#        for i in range(300):
#
#            output = network.forward(x_input) * current
#
#            s = torch.sum(soft_compare((output - p_input) * lambda2_)) / batch_size #disco rate
#            s2 = torch.sum(soft_compare((p_input - (mirror -output * fdr_scale)) * lambda2_)) / batch_size / float(fdr_scale) #false discoverate rate(over all samples)
#
#            if (s2/s).cpu().data[0] > alpha:
#                hi = current
#                current = (low + current)/2
#            else:
#                low = current
#                current = (hi + current)/2
#
#    print current, (s2/s).cpu().data[0]
#        
#    return current, (s2/s).cpu().data[0]
#
#import os
#import markdown2
#def generate_report(x = None, p = None, h = None, out_dir = '', url_prefix = '', info = {}, loss1 = None, loss2 = None, efdr = None, scales = None, x_prob = None, outputs = None, grids = None):
#    out_filename =  os.path.join(out_dir, 'report.html')
#    if not os.path.exists(out_dir):
#        os.makedirs(out_dir)
#    
#    if grids:
#        X_grid, Y_grid = grids
#    
#    f = open(out_filename, 'w')
#    s = ''
#    s += '# Report\n'
#    
#    s += '## Basic information\n'
#    for k in sorted(info.keys()):
#        v = info[k]
#        s += (str(k) + ':' + str(v) + '\n\n')
#    
#    dim = 1
#    if not x is None and not p is None :
#        n_samples = len(x)
#        s += '## Data description\n'
#        s += '- number of samples: {}\n'.format((n_samples))
#        s += '- dimension of data: {}\n'.format((x.shape[1]))
#        
#        dim = x.shape[1]
#        
#        plt.figure()
#        plt.scatter(x[:,0],p, 1, alpha = 0.3)
#        plt.xlabel('x')
#        plt.ylabel('p-value')
#        plt.savefig(out_dir + '/data.png')
#
#        s +=  '\n'
#        s += '![]({})\n'.format('data.png')
#        
#        if not h is None:
#            if not np.isnan(h[0]):
#                plt.figure()
#                plt.scatter(x[h==0,0], p[h==0], 1, alpha = 0.3, label = 'Null')
#                plt.scatter(x[h==1,0], p[h==1], 1, alpha = 0.3, label = 'Alternative')
#                plt.legend()
#                plt.xlabel('x')
#                plt.ylabel('p-value')
#                plt.savefig(out_dir + '/data_gt.png')
#        
#                s +=  '\n'
#                s += '![]({})\n'.format('data_gt.png')
#    
#    if not p is None:
#        plt.figure()
#        plt.hist(p, 50)
#        plt.xlabel('pvalues')
#        plt.savefig(out_dir + '/p_hist.png')
#        s +=  '\n'
#        s += '![]({})\n'.format('p_hist.png')
#    
#    if not loss1 is None and not loss2 is None:
#        s += '## Loss history\n'
#        plt.figure()
#        for i in range(len(loss1)):
#            plt.plot(np.log(loss1[i]), label = 'Loss history {}'.format(i))
#        plt.legend()
#        plt.savefig(out_dir + '/loss1.png')
#        s +=  '\n'
#        s += '![]({})\n'.format('loss1.png')
#        
#        plt.figure()
#        for i in range(len(loss2)):
#            plt.plot(loss2[i], label = 'Loss history {}'.format(i))
#        plt.legend()
#        plt.savefig(out_dir + '/loss2.png')
#        s +=  '\n'
#        s += '![]({})\n'.format('loss2.png')
#    
#    
#    if not scales is None and not efdr is None:
#        s += '## Scaling factors and expected FDR during training\n\n'
#        
#        s += str(efdr) + '\n\n'
#        
#        s += str(scales) + '\n\n'
#    
#    
#    if not len(outputs) == 0:
#        s += '## Threshold \n\n'
#        if dim == 1:
#            plt.figure()
#            for i in range(len(outputs)):
#                plt.plot(x_prob, outputs[i], label = 'threshold {}'.format(i))
#            
#            plt.xlabel('x')
#            plt.ylabel('t')
#            plt.legend()
#            plt.savefig(out_dir + '/threshold.png')
#
#            
#            s += '![]({})\n'.format('threshold.png')
#        elif dim == 2:
#            for i in range(len(outputs)):
#                plt.figure()
#                z = outputs[i].reshape(X_grid.shape)
#                plt.pcolor(X_grid, Y_grid, z)
#                plt.colorbar()
#                plt.savefig(out_dir + '/threshold_{}.png'.format(i))
#                s += '![]({})\n\n'.format('threshold_{}.png'.format(i))
#            
#     
#    html = markdown2.markdown(s)
#    
#    f.write(html)
#    f.close()
#    url = url_prefix + '/' + out_dir + '/report.html'
#    return url





"""Unit test for functions in util.py
"""
import numpy as np
import adafdr.util as util

def test_rank():
    """ Test for util.rank
    """
    x_test = np.array([[0.1, 2, 0.1, 5],
                       [0.1, 10, 5, 18]])
    x_test = x_test.T
    x_rank_continous = util.rank(x_test, continous_rank=True)
    x_rank_discrete = util.rank(x_test, continous_rank=False)
    print(x_rank_continous)
    assert all(x_rank_continous[:, 0] == [0, 2, 1, 3])
    print(x_rank_discrete)
    assert all(x_rank_discrete[:, 0] == [0.5, 2, 0.5, 3])
    


import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install
from subprocess import check_call
from distutils.command.build_py import build_py as _build_py
import sys, os.path

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="adafdr",
    version="0.0.9",
    author="Martin Zhang, Fei Xia, James Zou",
    author_email="jinye@stanford.edu",
    description="A fast and covariate-adaptive method for multiple hypothesis testing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/martinjzhang/adafdr",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    install_requires=[
       'numpy',
       'scipy',
       'scikit-learn',
       'torch==0.3.1',
       # 'multiprocessing',
       # 'logging',
       'matplotlib',
       # 'time',
       # 'os',
       ],
)
# setup(name='adafdr',
#       version='0.1.0',
#       description='A fast and covariate-adaptive method for multiple hypothesis testing.',
#       url='https://github.com/martinjzhang/adafdr',
#       author='Martin Zhang, Fei Xia, James Zou',
#       author_email='jinye@stanford.edu',
#       long_description=long_description,
#       long_description_content_type="text/markdown",
#       license='Stanford University',
#       packages=['adafdr'],
#       install_requires=[
#           'numpy',
#           'scipy',
#           'sklearn',
#           'torch',
#           # 'multiprocessing',
#           'logging',
#           'matplotlib',
#           # 'time',
#           # 'os',
#       ],
#       zip_safe=False)


import numpy as np
import scipy as sp

from adafdr.util import *
import adafdr.method as md
import adafdr.data_loader as dl

def test_f_slope():
    ''' a 1d example '''
    n_grid = 101
    x = get_grid_1d(n_grid)
    a = np.array([2])
    fx = dl.f_slope(x,a)
    p_true = np.exp(x*a).flatten()
    p_true = p_true / np.sum(p_true)    
    print('1d output dimension check:',fx.shape)
    assert fx.shape == (n_grid,)
    error = np.linalg.norm(fx/n_grid-p_true,1)
    print('1d n_grid normalize l1 error = %0.8f'%error)
    assert error<0.005    
    error = np.linalg.norm(fx/np.sum(fx)-p_true,1)
    print('1d self normalize l1 error = %0.8f'%error)
    assert error<1e-8
    
    ''' a 2d example '''
    n_grid = 101
    x = get_grid_2d(n_grid)
    a = np.array([2,0])
    fx = dl.f_slope(x,a)
    p_true = np.prod(np.exp(x*a),axis=1)
    p_true = p_true / np.sum(p_true)
    error = np.linalg.norm(fx/n_grid**2-p_true,1)
    print('2d output dimension check:',fx.shape)
    assert fx.shape == (n_grid**2,)
    print('2d l1 error = %0.8f'%error)
    assert error<0.005
    error = np.linalg.norm(fx/np.sum(fx)-p_true,1)
    print('2d self normalize l1 error = %0.8f'%error)
    assert error<1e-8
    
def test_ML_slope():
    np.random.seed(0)
    def get_l(a,x,c,v=None):
        if v is None:
            l = np.sum(np.log(a[a!=0]/(np.exp(a[a!=0])-1))) + \
                a.dot(np.mean(x,axis=0)) - c*np.sum(a**2)
        else:
            l = np.sum(np.log(a[a!=0]/(np.exp(a[a!=0])-1))) + \
                a.dot(np.sum((x.T*v).T,axis=0))/np.sum(v) - c*np.sum(a**2)
        return l
    
    x,param = dl.load_x_mixture(opt=0)
    a_true = param 
    c = 0
    
    ''' without weights '''
    a_hat = md.ML_slope(x,c=c)    
    error = np.linalg.norm(a_hat-a_true,1)
    print('without weights: a_true = %s'%a_true)
    print('without weights: a_hat = %s'%a_hat)
    print('without weights: a_error = %s\n'%error)
    assert error < 0.2
    
    l_hat = get_l(a_hat,x,c)
    l_true = get_l(a_true,x,c)
    error = np.absolute(l_hat-l_true)
    print('without weights: l_true = %s'%l_true)
    print('without weights: l_hat = %s'%l_hat)
    print('without weights: l_error = %s\n'%error)    
    assert error < 0.001
    
    ''' with weights '''
    v = np.random.randn(x.shape[0]).clip(min=0)  
    a_hat = md.ML_slope(x,v=v,c=c)
    error = np.linalg.norm(a_hat-a_true,1)
    print('with weights: a_true = %s'%a_true)
    print('with weights: a_hat = %s'%a_hat)
    print('with weights: a_error = %s\n'%error)
    assert error < 0.2
    
    l_hat = get_l(a_hat,x,c,v)
    l_true = get_l(a_true,x,c,v)
    error = np.absolute(l_hat-l_true)
    print('with weights: l_true = %s'%l_true)
    print('with weights: l_hat = %s'%l_hat)
    print('with weights: l_error = %s\n'%error)    
    assert error < 0.001
        
def test_f_bump():
    ''' a 1d example '''
    n_grid = 101
    x = get_grid_1d(n_grid)
    mu = np.array([0.5],dtype=float)
    sigma = np.array([0.1],dtype=float)    
    fx = dl.f_bump(x,mu,sigma)    
    p_true = np.exp(-(x-mu)**2/sigma**2/2).reshape(-1)
    p_true = p_true / np.sum(p_true)    
    print('1d output dimension check:',fx.shape)
    assert fx.shape == (n_grid,)
    error = np.linalg.norm(fx/n_grid-p_true,1)
    print('1d n_grid normalize l1 error = %0.8f'%error)
    assert error<0.01  
    error = np.linalg.norm(fx/np.sum(fx)-p_true,1)
    print('1d self normalize l1 error = %0.8f'%error)
    assert error<1e-8
    
    ''' a 2d example '''
    n_grid = 101
    x = get_grid_2d(n_grid)
    mu = np.array([0.5,0.2],dtype=float)
    sigma = np.array([0.1,0.1],dtype=float)    
    fx = dl.f_bump(x,mu,sigma)    
    p_true = np.exp(-np.sum((x-mu)**2/sigma**2/2,axis=1)).reshape(-1)
    p_true = p_true / np.sum(p_true)    
    print('2d output dimension check:',fx.shape)
    assert fx.shape == (n_grid**2,)
    print('2d l1 error = %0.8f'%error)
    assert error<0.01
    error = np.linalg.norm(fx/np.sum(fx)-p_true,1)
    print('2d self normalize l1 error = %0.8f'%error)
    assert error<1e-8
    
def test_ML_slope():
    np.random.seed(0)
    def get_l(param,x,v=None): # omitting the constant term 
        l = 0
        mu,sigma = param
        for j in range(x.shape[1]):        
            Z = sp.stats.norm.cdf(1,loc=mu[j],scale=sigma[j])-sp.stats.norm.cdf(0,loc=mu[j],scale=sigma[j])
            print(Z)

            if v is None:
                l += -np.log(Z) - np.log(sigma[j]) - 1/2/sigma[j]**2*np.mean((x[:,j]-mu[j])**2)
            else:
                t = np.sum((x[:,j]-mu[j])**2*v) / np.sum(v)
                l += -np.log(Z) - np.log(sigma[j]) - 1/2/sigma[j]**2*t - 1/2*np.log(2*np.pi)
        return l

    x,param = dl.load_x_mixture(opt=1)
    mu_true,sigma_true = param
    
    ''' without weights '''
    mu_hat,sigma_hat = md.ML_bump(x)
    error = np.linalg.norm(mu_true-mu_hat,1) + np.linalg.norm(sigma_true-sigma_hat,1)
    print('without weights: mu_true = %s, sigma_true = %s'%(mu_true,sigma_true))
    print('without weights: mu_hat = %s, sigma_hat = %s'%(mu_hat,sigma_hat))
    print('without weights: param_error = %s\n'%error)
    assert error < 0.2

    l_hat = get_l((mu_hat,sigma_hat),x)
    l_true = get_l((mu_true,sigma_true),x)
    error = np.absolute(l_hat-l_true)
    print('without weights: l_true = %s'%l_true)
    print('without weights: l_hat = %s'%l_hat)
    print('without weights: l_error = %s\n'%error)    
    assert error < 0.01

    ''' with weights '''
    v = np.random.randn(x.shape[0]).clip(min=0)  
    mu_hat,sigma_hat = md.ML_bump(x,v=v)    
    error = np.linalg.norm(mu_true-mu_hat,1) + np.linalg.norm(sigma_true-sigma_hat,1)
    print('with weights: mu_true = %s, sigma_true = %s'%(mu_true,sigma_true))
    print('with weights: mu_hat = %s, sigma_hat = %s'%(mu_hat,sigma_hat))
    print('with weights: param_error = %s\n'%error)
    assert error < 0.2

    l_hat = get_l((mu_hat,sigma_hat),x,v=v)
    l_true = get_l((mu_true,sigma_true),x,v=v)
    error = np.absolute(l_hat-l_true)
    print('with weights: l_true = %s'%l_true)
    print('with weights: l_hat = %s'%l_hat)
    print('with weights: l_error = %s\n'%error)    
    assert error < 0.01  
    # assert False
    
def test_f_all():
    n_grid = 101
    x = get_grid_2d(n_grid)
    w = np.array([0.4,0.3,0.3],dtype=float)
    a = np.array([2,0],dtype=float)
    mu = np.array([[0.2,0.2],[0.7,0.7]],dtype=float)
    sigma = np.array([[0.1,0.2],[0.1,0.1]],dtype=float) 
    
    fx = dl.f_all(x,a,mu,sigma,w)    
    fx0 = md.f_slope(x,a)
    fx1 = md.f_bump(x,mu[0],sigma[0])
    fx2 = md.f_bump(x,mu[1],sigma[1])
    error = np.linalg.norm(fx - (w[0]*fx0 + w[1]*fx1 + w[2]*fx2),1)
    print('theoretical error = %0.8f'%error)
    
    
    p0 = np.prod(np.exp(x*a),axis=1)
    p0 /= p0.sum()
    p1 = np.exp(-np.sum((x-mu[0])**2/sigma[0]**2/2,axis=1)).reshape(-1)
    p1 /= p1.sum()
    p2 = np.exp(-np.sum((x-mu[1])**2/sigma[1]**2/2,axis=1)).reshape(-1)
    p2 /= p2.sum() 
    p_true = w[0]*p0 + w[1]*p1 + w[2]*p2
    error = np.linalg.norm(fx/n_grid**2-p_true,1)
    print('output dimension check:',fx.shape)
    assert fx.shape == (n_grid**2,)
    print('all l1 error = %0.8f'%error)
    assert error<0.01
    
def test_mixture_fit():
    def get_l(x,a,mu,sigma,w):
        fx = md.f_all(x,a,mu,sigma,w)
        return np.mean(np.log(fx))
    # 2d slope+bump
    x,param = dl.load_x_mixture(opt=2)
    a_true,mu_true,sigma_true,w_true = param
    l_true = get_l(x,a_true,mu_true,sigma_true,w_true)
    a_hat,mu_hat,sigma_hat,w_hat = md.mixture_fit(x)
    l_hat = get_l(x,a_hat,mu_hat,sigma_hat,w_hat)
    print('# 2d l_true=%s'%l_true)
    print('# 2d l_hat=%s'%l_hat)   
    error = np.absolute(l_hat-l_true)
    assert error<0.01
    
    # 10d slope+bump
    x,param = dl.load_x_mixture(opt=3)
    a_true,mu_true,sigma_true,w_true = param
    l_true = get_l(x,a_true,mu_true,sigma_true,w_true)
    a_hat,mu_hat,sigma_hat,w_hat = md.mixture_fit(x)
    l_hat = get_l(x,a_hat,mu_hat,sigma_hat,w_hat)
    print('# 10d l_true=%s'%l_true)
    print('# 10d l_hat=%s'%l_hat)   
    error = np.absolute(l_hat-l_true)
    assert error<0.01
